import factory

from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType

from .models import Article, UserObjectPermissions


def _get_perm(perm_name):
    """
    Returns permission instance with given name.

    Permission name is a string like 'auth.add_user'.
    """
    app_label, codename = perm_name.split(".")
    return Permission.objects.get(
        content_type__app_label=app_label, codename=codename
    )


class ArticleFactory(factory.django.DjangoModelFactory):
    """Generates Articles"""
    title = factory.Sequence(lambda n: f"Article number {n}")
    body = factory.Sequence(lambda n: "Body of article {n}")

    class Meta:
        model = Article
        abstract = False


class GroupFactory(factory.django.DjangoModelFactory):
    """Artificial divides as a service"""
    name = factory.Sequence(lambda n: f"group{n}")

    class Meta:
        model = Group
        abstract = False


class UserFactory(factory.django.DjangoModelFactory):
    """The people who make it all possible"""
    username = factory.Sequence(lambda n: f"user{n}")
    first_name = factory.Sequence(lambda n: f"John {n}")
    last_name = factory.Sequence(lambda n: f"Doe {n}")
    email = factory.Sequence(lambda n: f"user{n}@example.com")
    password = factory.PostGenerationMethodCall("set_password", "asdf1234")

    class Meta:
        model = User
        abstract = False

    @factory.post_generation
    def permissions(self, create, extracted, **kwargs):
        """Give the user some permissions"""
        if create and extracted:
            # We have a saved object and a list of permission names
            self.user_permissions.add(*[_get_perm(pn) for pn in extracted])


class ContentTypeFactory(factory.django.DjangoModelFactory):
    """Factory for creating `ContentType` model objects"""
    app_label = factory.Sequence(lambda n: f"app_label_{n}")
    model = factory.Sequence(lambda n: f"model_{n}")

    class Meta:
        model = ContentType
        abstract = False


class PermissionFactory(factory.django.DjangoModelFactory):
    """Factory for creating `Permission` model objects"""
    name = factory.Sequence(lambda n: f"name_{n}")
    codename = factory.Sequence(lambda n: f"codename_{n}")
    content_type = factory.SubFactory(ContentTypeFactory)

    class Meta:
        model = Permission
        abstract = False


class UserObjectPermissionsFactory(factory.django.DjangoModelFactory):
    """Factory for creating `UserObjectPermissions` model objects"""
    user = factory.SubFactory(UserFactory)
    permission = factory.SubFactory(PermissionFactory)
    article_object = factory.SubFactory(ArticleFactory)

    class Meta:
        model = UserObjectPermissions
        abstract = False



import json
from unittest import mock

from django import test
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django.utils.encoding import force_str

from braces.views import AjaxResponseMixin

from .factories import ArticleFactory, UserFactory
from .helpers import TestViewHelper
from .views import SimpleJsonView, JsonRequestResponseView, JsonBadRequestView


class TestAjaxResponseMixin(TestViewHelper, test.TestCase):
    """
    Tests for AjaxResponseMixin.
    """

    methods = ["get", "post", "put", "delete"]

    def test_xhr(self):
        """
        Checks if ajax_* method has been called for every http method.
        """
        # AjaxResponseView returns 'AJAX_OK' when requested with XmlHttpRequest
        for m in self.methods:
            fn = getattr(self.client, m)
            resp = fn(
                "/ajax_response/", HTTP_X_REQUESTED_WITH="XMLHttpRequest"
            )
            assert force_str(resp.content) == "AJAX_OK"

    def test_not_xhr(self):
        """
        Normal methods (get, post, etc) should be used when handling non-ajax
        requests.
        """
        for m in self.methods:
            fn = getattr(self.client, m)
            resp = fn("/ajax_response/")
            assert force_str(resp.content) == "OK"

    def test_fallback_to_normal_methods(self):
        """
        Ajax methods should fallback to normal methods by default.
        """
        test_cases = [
            ("get", "get"),
            ("post", "post"),
            ("put", "get"),
            ("delete", "get"),
        ]

        for ajax_method, fallback in test_cases:
            m, mixin = mock.Mock(), AjaxResponseMixin()
            m.return_value = HttpResponse()
            req = self.build_request()
            setattr(mixin, fallback, m)
            fn = getattr(mixin, "{0}_ajax".format(ajax_method))
            ret = fn(req, 1, 2, meth=ajax_method)
            # check if appropriate method has been called
            m.assert_called_once_with(req, 1, 2, meth=ajax_method)
            # check if appropriate value has been returned
            self.assertIs(m.return_value, ret)


class TestJSONResponseMixin(TestViewHelper, test.TestCase):
    """
    Tests for JSONResponseMixin.
    """

    view_class = SimpleJsonView

    def assert_json_response(self, resp, status_code=200):
        self.assertEqual(status_code, resp.status_code)
        self.assertEqual(
            "application/json", resp["content-type"].split(";")[0]
        )

    def get_content(self, url):
        """
        GET url and return content
        """
        resp = self.client.get(url)
        self.assert_json_response(resp)
        content = force_str(resp.content)
        return content

    def test_simple_json(self):
        """
        Tests render_json_response() method.
        """
        user = UserFactory()
        self.client.login(username=user.username, password="asdf1234")
        data = json.loads(self.get_content("/simple_json/"))
        self.assertEqual({"username": user.username}, data)

    def test_serialization(self):
        """
        Tests render_json_object_response() method which serializes objects
        using django's serializer framework.
        """
        a1, a2 = [ArticleFactory() for __ in range(2)]
        data = json.loads(self.get_content("/article_list_json/"))
        self.assertIsInstance(data, list)
        self.assertEqual(2, len(data))
        titles = []
        for row in data:
            # only title has been serialized
            self.assertEqual(1, len(row["fields"]))
            titles.append(row["fields"]["title"])

        self.assertIn(a1.title, titles)
        self.assertIn(a2.title, titles)

    def test_bad_content_type(self):
        """
        ImproperlyConfigured exception should be raised if content_type
        attribute is not set correctly.
        """
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(self.build_request(), content_type=["a"])

    def test_pretty_json(self):
        """
        Success if JSON responses are the same, and the well-indented response
        is longer than the normal one.
        """
        user = UserFactory()
        self.client.login(username=user.username, password="asfa")
        normal_content = self.get_content("/simple_json/")
        self.view_class.json_dumps_kwargs = {"indent": 2}
        pretty_content = self.get_content("/simple_json/")
        normal_json = json.loads("{0}".format(normal_content))
        pretty_json = json.loads("{0}".format(pretty_content))
        self.assertEqual(normal_json, pretty_json)
        self.assertTrue(len(pretty_content) > len(normal_content))

    def test_json_encoder_class_atrribute(self):
        """
        Tests setting custom `json_encoder_class` attribute.
        """
        data = json.loads(self.get_content("/simple_json_custom_encoder/"))
        self.assertEqual({"numbers": [1, 2, 3]}, data)


class TestJsonRequestResponseMixin(TestViewHelper, test.TestCase):
    view_class = JsonRequestResponseView
    request_dict = {"status": "operational"}

    def test_get_request_json_properly_formatted(self):
        """
        Properly formatted JSON requests should result in a JSON object
        """
        data = json.dumps(self.request_dict).encode("utf-8")
        response = self.client.post(
            "/json_request/", content_type="application/json", data=data
        )
        response_json = json.loads(response.content.decode("utf-8"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response_json, self.request_dict)

    def test_get_request_json_improperly_formatted(self):
        """
        Improperly formatted JSON requests should make request_json == None
        """
        response = self.client.post("/json_request/", data=self.request_dict)
        response_json = json.loads(response.content.decode("utf-8"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response_json, None)

    def test_bad_request_response_with_custom_error_message(self):
        """
        If a view calls render_bad_request_response when request_json is empty
        or None, the client should get a 400 error
        """
        response = self.client.post(
            "/json_custom_bad_request/", data=self.request_dict
        )
        response_json = json.loads(response.content.decode("utf-8"))
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response_json, {"error": "you messed up"})


class TestJsonBadRequestMixin(TestViewHelper, test.TestCase):
    view_class = JsonBadRequestView
    request_dict = {"status": "operational"}

    def test_bad_request_response(self):
        """
        If a view calls render_bad_request_response when request_json is empty
        or None, the client should get a 400 error
        """
        response = self.client.post(
            "/json_bad_request/", data=self.request_dict
        )
        response_json = json.loads(response.content.decode("utf-8"))
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response_json, self.view_class.error_response_dict)

    def test_options_request_with_required_json_should_pass(self):
        """
        If a the client sends an OPTIONS request,
        even if require_json is set to true
        the client should get not a 400 error, because:
        * it's not possible to send HTTP body within an OPTIONS request
        * it's not up to an OPTIONS request to decide if the payload is valid
        """
        response = self.client.options("/json_bad_request/", data=None)
        self.assertEqual(response.status_code, 200)



from unittest import mock
import pytest

from django.core.exceptions import ImproperlyConfigured
from django import test

from .helpers import TestViewHelper
from .models import Article
from .views import (
    ArticleListView,
    ArticleListViewWithCustomQueryset,
    AuthorDetailView,
    OrderableListView
)


class TestSelectRelatedMixin(TestViewHelper, test.TestCase):
    """Scenarios related to adding select_related to queries"""
    view_class = ArticleListView

    def test_missing_select_related(self):
        """If select_related is unset, raise ImproperlyConfigured"""
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(self.build_request(), select_related=None)

    def test_invalid_select_related(self):
        """If select_related is not a list or tuple, raise ImproperlyConfigured"""
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(self.build_request(), select_related={"a": 1})

    @mock.patch("django.db.models.query.QuerySet.select_related")
    def test_select_related_called(self, m):
        """QuerySet.select_related should be called with the correct arguments"""
        qs = Article.objects.all()
        m.return_value = qs.select_related("author")
        qs.select_related = m
        m.reset_mock()

        resp = self.dispatch_view(self.build_request())
        self.assertEqual(200, resp.status_code)
        m.assert_called_once_with("author")

    @mock.patch("django.db.models.query.QuerySet.select_related")
    def test_select_related_keeps_select_related_from_queryset(self, m):
        """
        Checks that an empty select_related attribute does not
        cancel a select_related provided by queryset.
        """
        qs = Article.objects.all()
        qs.select_related = m
        m.reset_mock()

        with pytest.warns(UserWarning):
            resp = self.dispatch_view(
                self.build_request(),
                view_class=ArticleListViewWithCustomQueryset,
            )
        self.assertEqual(200, resp.status_code)
        self.assertEqual(0, m.call_count)


class TestPrefetchRelatedMixin(TestViewHelper, test.TestCase):
    """Scenarios related to adding prefetch_related to queries"""
    view_class = AuthorDetailView

    def test_missing_prefetch_related(self):
        """If prefetch_related is missing/None, raise ImproperlyConfigured"""
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(self.build_request(), prefetch_related=None)

    def test_invalid_prefetch_related(self):
        """If prefetch_related is not a list or tuple, raise ImproperlyConfigured"""
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(self.build_request(), prefetch_related={"a": 1})

    @mock.patch("django.db.models.query.QuerySet.prefetch_related")
    def test_prefetch_related_called(self, m):
        """QuerySet.prefetch_related() should be called with correct arguments"""
        qs = Article.objects.all()
        m.return_value = qs.prefetch_related("article_set")
        qs.prefetch_related = m
        m.reset_mock()

        resp = self.dispatch_view(self.build_request())
        self.assertEqual(200, resp.status_code)
        m.assert_called_once_with("article_set")

    @mock.patch("django.db.models.query.QuerySet.prefetch_related")
    def test_prefetch_related_keeps_select_related_from_queryset(self, m):
        """
        Checks that an empty prefetch_related attribute does not
        cancel a prefetch_related provided by queryset.
        """
        qs = Article.objects.all()
        qs.prefetch_related = m
        m.reset_mock()

        with pytest.warns(UserWarning):
            resp = self.dispatch_view(
                self.build_request(),
                view_class=ArticleListViewWithCustomQueryset,
            )
        self.assertEqual(200, resp.status_code)
        self.assertEqual(0, m.call_count)


class TestOrderableListMixin(TestViewHelper, test.TestCase):
    """Scenarios involving ordering records"""
    view_class = OrderableListView

    def __make_test_articles(self):
        """Generate a couple of articles"""
        a1 = Article.objects.create(title="Alpha", body="Zet")
        a2 = Article.objects.create(title="Zet", body="Alpha")
        return a1, a2

    def test_correct_order(self):
        """Valid column and order query arguments should order the objects"""
        a1, a2 = self.__make_test_articles()

        resp = self.dispatch_view(
            self.build_request(path="?order_by=title&ordering=asc"),
            orderable_columns=None,
            get_orderable_columns=lambda: (
                "id",
                "title",
            ),
        )
        self.assertEqual(list(resp.context_data["object_list"]), [a1, a2])

        resp = self.dispatch_view(
            self.build_request(path="?order_by=id&ordering=desc"),
            orderable_columns=None,
            get_orderable_columns=lambda: (
                "id",
                "title",
            ),
        )
        self.assertEqual(list(resp.context_data["object_list"]), [a2, a1])

    def test_correct_order_with_default_ordering(self):
        """A valid order_by query argument should sort the default direction"""
        a1, a2 = self.__make_test_articles()

        resp = self.dispatch_view(
            self.build_request(path="?order_by=id"),
            orderable_columns=None,
            ordering_default=None,
            get_orderable_columns=lambda: (
                "id",
                "title",
            ),
        )
        self.assertEqual(list(resp.context_data["object_list"]), [a1, a2])

        resp = self.dispatch_view(
            self.build_request(path="?order_by=id"),
            orderable_columns=None,
            ordering_default="asc",
            get_orderable_columns=lambda: (
                "id",
                "title",
            ),
        )
        self.assertEqual(list(resp.context_data["object_list"]), [a1, a2])

        resp = self.dispatch_view(
            self.build_request(path="?order_by=id"),
            orderable_columns=None,
            ordering_default="desc",
            get_orderable_columns=lambda: (
                "id",
                "title",
            ),
        )
        self.assertEqual(list(resp.context_data["object_list"]), [a2, a1])

    def test_correct_order_with_param_not_default_ordering(self):
        """
        Objects must be properly ordered if requested with valid column names
        and ordering option in the query params.
        In this case, the ordering_default will be overwritten.
        """
        a1, a2 = self.__make_test_articles()

        resp = self.dispatch_view(
            self.build_request(path="?order_by=id&ordering=asc"),
            orderable_columns=None,
            ordering_default="desc",
            get_orderable_columns=lambda: (
                "id",
                "title",
            ),
        )
        self.assertEqual(list(resp.context_data["object_list"]), [a1, a2])

    def test_correct_order_with_incorrect_default_ordering(self):
        """
        Objects must be properly ordered if requested with valid column names
        and with the default ordering
        """
        view = self.view_class()
        view.ordering_default = "improper_default_value"
        self.assertRaises(
            ImproperlyConfigured, lambda: view.get_ordering_default()
        )

    def test_default_column(self):
        """
        When no ordering specified in GET, use
        View.get_orderable_columns_default()
        """
        a1, a2 = self.__make_test_articles()

        resp = self.dispatch_view(self.build_request())
        self.assertEqual(list(resp.context_data["object_list"]), [a1, a2])

    def test_get_orderable_columns_returns_correct_values(self):
        """
        OrderableListMixin.get_orderable_columns() should return
        View.orderable_columns attribute by default or raise
        ImproperlyConfigured exception if the attribute is None
        """
        view = self.view_class()
        self.assertEqual(view.get_orderable_columns(), view.orderable_columns)
        view.orderable_columns = None
        self.assertRaises(
            ImproperlyConfigured, lambda: view.get_orderable_columns()
        )

    def test_get_orderable_columns_default_returns_correct_values(self):
        """
        OrderableListMixin.get_orderable_columns_default() should return
        View.orderable_columns_default attribute by default or raise
        ImproperlyConfigured exception if the attribute is None
        """
        view = self.view_class()
        self.assertEqual(
            view.get_orderable_columns_default(),
            view.orderable_columns_default,
        )
        view.orderable_columns_default = None
        self.assertRaises(
            ImproperlyConfigured, lambda: view.get_orderable_columns_default()
        )

    def test_only_allowed_columns(self):
        """
        If column is not in Model.Orderable.columns iterable, the objects
        should be ordered by default column.
        """
        a1, a2 = self.__make_test_articles()

        resp = self.dispatch_view(
            self.build_request(path="?order_by=body&ordering=asc"),
            orderable_columns_default=None,
            get_orderable_columns_default=lambda: "title",
        )
        self.assertEqual(list(resp.context_data["object_list"]), [a1, a2])



import json
from django.core import serializers
from django.core.exceptions import ImproperlyConfigured
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse, HttpResponseBadRequest, JsonResponse


class JSONResponseMixin:
    """
    Basic serialized responses.

    For anything more complicated than basic Python types or Django
    models, please use something like django-rest-framework.
    """

    content_type = None
    json_dumps_kwargs = None
    json_encoder_class = None

    def get_content_type(self):
        """Get the appropriate content type for the response"""
        if self.content_type is not None and not isinstance(
            self.content_type, str
        ):
            class_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{class_name} is missing a content type. Define {class_name}"
                ".content_type or override {class_name}.get_content_type()."
            )
        return self.content_type or "application/json"

    def get_json_dumps_kwargs(self):
        """Get kwargs for custom JSON compilation"""
        dumps_kwargs = getattr(self, "json_dumps_kwargs", None) or {}
        dumps_kwargs.setdefault("ensure_ascii", False)
        return dumps_kwargs

    def get_json_encoder_class(self):
        """Get the encoder class to use"""
        if self.json_encoder_class is None:
            self.json_encoder_class = DjangoJSONEncoder
        return self.json_encoder_class

    def render_json_response(self, context_dict, status=200):
        """
        Limited serialization for shipping plain data.
        Do not use for models or other complex objects.
        """
        response = JsonResponse(
            data=context_dict,
            safe=False,
            encoder=self.get_json_encoder_class(),
            json_dumps_params=self.get_json_dumps_kwargs(),
            content_type=self.get_content_type(),
            status=status
        )
        return response

    def render_json_object_response(self, objects, **kwargs):
        """
        Serializes objects using Django's builtin JSON serializer. Additional
        kwargs can be used the same way for django.core.serializers.serialize.
        """
        try:
            response = self.render_json_response(objects, **kwargs)
        except TypeError:
            json_data = serializers.serialize("json", objects, **kwargs)
            response = HttpResponse(json_data, content_type=self.get_content_type())
        return response


class AjaxResponseMixin:
    """
    Mixin allows you to define alternative methods for ajax requests. Similar
    to the normal get, post, and put methods, you can use get_ajax, post_ajax,
    and put_ajax.
    """

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate handler method"""
        if all([
            request.headers.get("x-requested-with") == "XMLHttpRequest",
            request.method.lower() in self.http_method_names
        ]):
            handler = getattr(
                self,
                f"{request.method.lower()}_ajax",
                self.http_method_not_allowed,
            )
            self.request = request
            self.args = args
            self.kwargs = kwargs
            return handler(request, *args, **kwargs)

        return super().dispatch(request, *args, **kwargs)

    def get_ajax(self, request, *args, **kwargs):
        """Handle a GET request made with AJAX"""
        return self.get(request, *args, **kwargs)

    def post_ajax(self, request, *args, **kwargs):
        """Handle a POST request made with AJAX"""
        return self.post(request, *args, **kwargs)

    def put_ajax(self, request, *args, **kwargs):
        """Handle a PUT request made with AJAX"""
        return self.get(request, *args, **kwargs)

    def delete_ajax(self, request, *args, **kwargs):
        """Handle a DELETE request made with AJAX"""
        return self.get(request, *args, **kwargs)


class JsonRequestResponseMixin(JSONResponseMixin):
    """
    Attempt to parse the request body as JSON.

    If successful, self.request_json will contain the deserialized object.
    Otherwise, self.request_json will be None.

    Set the attribute require_json to True to return a 400 "Bad Request" error
    for requests that don't contain JSON.

    Note: To allow public access to your view, you'll need to use the
    csrf_exempt decorator or CsrfExemptMixin.

    Example Usage:

        class SomeView(CsrfExemptMixin, JsonRequestResponseMixin):
            def post(self, request, *args, **kwargs):
                do_stuff_with_contents_of_request_json()
                return self.render_json_response(
                    {'message': 'Thanks!'})
    """

    require_json = False
    error_response_dict = {"errors": ["Improperly formatted request"]}

    def render_bad_request_response(self, error_dict=None):
        """Generate errors for bad content"""
        if error_dict is None:
            error_dict = self.error_response_dict
        json_context = json.dumps(
            error_dict,
            cls=self.get_json_encoder_class(),
            **self.get_json_dumps_kwargs()
        ).encode("utf-8")
        return HttpResponseBadRequest(
            json_context, content_type=self.get_content_type()
        )

    def get_request_json(self):
        """Get the JSON included in the body"""
        try:
            return json.loads(self.request.body.decode("utf-8"))
        except (json.JSONDecodeError, ValueError):
            return None

    def dispatch(self, request, *args, **kwargs):
        """Trigger the appropriate method"""
        self.request = request
        self.args = args
        self.kwargs = kwargs

        self.request_json = self.get_request_json()
        if all(
            [
                request.method != "OPTIONS",
                self.require_json,
                self.request_json is None,
            ]
        ):
            return self.render_bad_request_response()
        return super().dispatch(request, *args, **kwargs)


class JSONRequestResponseMixin(JsonRequestResponseMixin):
    """Convenience alias"""



from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect
from django.views.decorators.cache import cache_control, never_cache
from django.urls import resolve
from django.utils.encoding import force_str
from django.urls import resolve


class SetHeadlineMixin:
    """
    Define a `headline` context item as a view attribute
    """

    headline = None  # Default the headline to none

    def get_context_data(self, **kwargs):
        """Add the headline to the context"""
        kwargs = super().get_context_data(**kwargs)
        kwargs.update({"headline": self.get_headline()})
        return kwargs

    def get_headline(self):
        """Fetch the headline from the instance"""
        if self.headline is None:
            class_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{class_name} is missing the headline attribute. "
                f"Define {class_name}.headline, or override {class_name}.get_headline()."
            )
        return force_str(self.headline)


class StaticContextMixin:
    """
    Set static context items via an attribute on the view.
    """

    static_context = None

    def get_context_data(self, **kwargs):
        """Update the context to include the static content"""
        kwargs = super().get_context_data(**kwargs)

        try:
            kwargs.update(self.get_static_context())
        except (TypeError, ValueError):
            raise ImproperlyConfigured(
                f"{self.__class__.__name__}.static_context must be a "
                "dictionary or a series of two-tuples."
            )
        else:
            return kwargs

    def get_static_context(self):
        """Fetch the static content from the view"""
        if self.static_context is None:
            class_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{class_name} is missing the static_context attribute. Define "
                f"{class_name}.static_context, or override {class_name}.get_static_context()"
            )
        return self.static_context


class CanonicalSlugDetailMixin:
    """
    Enforce a canonical slug in the URL.

    If a URL takes a object's pk and slug as arguments and the slug URL
    argument does not equal the object's canonical slug, this mixin will
    redirect to the URL containing the canonical slug.
    """

    def dispatch(self, request, *args, **kwargs):
        """
        Redirect to the appropriate URL if necessary.
        Otherwise, trigger HTTP-method-appropriate handler.
        """
        self.request = request
        self.args = args
        self.kwargs = kwargs

        # Get the current object, url slug, and url name.
        obj = self.get_object()
        slug = self.kwargs.get(self.slug_url_kwarg, None)
        match = resolve(request.path_info)
        url_parts = match.namespaces
        url_parts.append(match.url_name)
        current_urlpattern = ":".join(url_parts)

        # Find the canonical slug for the object
        if hasattr(obj, "get_canonical_slug"):
            canonical_slug = obj.get_canonical_slug()
        else:
            canonical_slug = self.get_canonical_slug()

        # Redirect if current slug is not the canonical one
        if canonical_slug != slug:
            params = {
                self.pk_url_kwarg: obj.pk,
                self.slug_url_kwarg: canonical_slug,
                "permanent": True,
            }
            return redirect(current_urlpattern, **params)

        return super().dispatch(request, *args, **kwargs)

    def get_canonical_slug(self):
        """
        Provide a method to return the correct slug for this object.
        """
        return self.get_object().slug


class AllVerbsMixin:
    """
    Call a single method for all HTTP verbs.

    The name of the method should be specified using the class attribute
    `all_handler`. The default value of this attribute is 'all'.
    """

    all_handler = "all"

    def dispatch(self, request, *args, **kwargs):
        """Call the all handler"""
        if not self.all_handler:
            raise ImproperlyConfigured(
                f"{self.__class__.__name__} requires the all_handler attribute to be set."
            )

        handler = getattr(self, self.all_handler, self.http_method_not_allowed)
        return handler(request, *args, **kwargs)


class HeaderMixin:
    """
    Add extra HTTP headers to a response by specifying them in the
    ``headers`` attribute or by overriding the ``get_headers()`` method.
    """

    headers = {}

    def get_headers(self, request):
        return self.headers

    def dispatch(self, request, *args, **kwargs):
        """
        Override this method to customize the way additional headers are
        retrieved.  It is mandatory that the returned value supports the
        ``.items()`` method.
        """
        response = super().dispatch(request, *args, **kwargs)
        for key, value in self.get_headers(request).items():
            if key not in response:
                response[key] = value
        return response


class CacheControlMixin:
    """
    Mixin that allows setting Cache-Control options.

    Specify Cache-Control options as class attributes on the view class.

    Cache-Control directive explanations:
    http://condor.depaul.edu/dmumaugh/readings/handouts/SE435/HTTP/node24.html

    Django's ``django.views.decorators.cache.cache_control`` options:
    https://docs.djangoproject.com/en/dev/topics/cache/#controlling-cache-using-other-headers
    """
    # These are all ``None``, which indicates unset.
    cachecontrol_public = None
    cachecontrol_private = None
    cachecontrol_no_cache = None
    cachecontrol_no_transform = None
    cachecontrol_must_revalidate = None
    cachecontrol_proxy_revalidate = None
    cachecontrol_max_age = None
    cachecontrol_s_maxage = None

    @classmethod
    def get_cachecontrol_options(cls):
        """Compile a dictionary of selected cache options"""
        opts = (
            'public', 'private', 'no_cache', 'no_transform',
            'must_revalidate', 'proxy_revalidate', 'max_age',
            's_maxage'
        )
        options = {}
        for opt in opts:
            value = getattr(cls, f'cachecontrol_{opt}', None)
            if value is not None:
                options[opt] = value
        return options

    @classmethod
    def as_view(cls, *args, **kwargs):
        """Wrap the view with appropriate cache controls"""
        view_func = super().as_view(*args, **kwargs)
        options = cls.get_cachecontrol_options()
        return cache_control(**options)(view_func)


class NeverCacheMixin:
    """
    Mixin that applies Django's `never_cache` view decorator to prevent
    upstream HTTP-based caching.
    """
    @classmethod
    def as_view(cls, *args, **kwargs):
        """
        Wrap the view with the `never_cache` decorator.
        """
        view_func = super().as_view(*args, **kwargs)
        return never_cache(view_func)



from functools import partial

from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.utils.decorators import method_decorator
from django.utils.encoding import force_str
from django.utils.functional import Promise
from django.views.decorators.csrf import csrf_exempt
from django.urls import reverse


class CsrfExemptMixin:
    """
    Exempts the view from CSRF requirements.

    NOTE:
        This should be the left-most mixin of a view.
    """

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)


class UserFormKwargsMixin:
    """
    Automatically include `request.user` in form kwargs.

    ## Note
    You will need to handle the `user` kwarg in your form. Usually
    this means `user = kwargs.pop("user")` in your form's `__init__`.
    """

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        # Update the existing form kwargs dict with the request's user.
        kwargs.update({"user": self.request.user})
        return kwargs


class SuccessURLRedirectListMixin:
    """
    Automatically reverses `success_list_url` and returns that as
    the `success_url` for a form view.

    This is meant to redirect to a view without arguments. If you need
    to include arguments to `reverse`, you can omit this mixin.
    """

    success_list_url = None  # Default the success url to none

    def get_success_url(self):
        # Return the reversed success url.
        if self.success_list_url is None:
            class_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{class_name} is missing a success_list_url attribute. "
                f"Define {class_name}.success_list_url or override "
                f"{class_name}.get_success_url()."
            )
        return reverse(self.success_list_url)


class _MessageAPIWrapper:
    """
    Wrapper for the django.contrib.messages.api module.
    Automatically pass a request object as the first parameter of
    message function calls.
    """

    API = set(
        [
            "add_message",
            "get_messages",
            "get_level",
            "set_level",
            "debug",
            "info",
            "success",
            "warning",
            "error",
        ]
    )

    def __init__(self, request):
        for name in self.API:
            api_fn = getattr(messages.api, name)
            setattr(self, name, partial(api_fn, request))


class _MessageDescriptor:
    """
    A descriptor that binds the _MessageAPIWrapper to the view's
    request.
    """

    def __get__(self, instance, *args, **kwargs):
        return _MessageAPIWrapper(instance.request)


class MessageMixin:
    """
    Add a `messages` attribute on the view instance that wraps
    `django.contrib.messages`, automatically passing the current
    request object.
    """

    messages = _MessageDescriptor()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._class_name = self.__class__.__name__


class FormValidMessageMixin(MessageMixin):
    """
    Set a string to be sent via Django's messages framework when a form
    passes validation.
    """

    form_valid_message = None  # Default to None

    def get_form_valid_message(self):
        """
        Validate that form_valid_message is set correctly
        """
        if self.form_valid_message is None:
            raise ImproperlyConfigured(
                f"{self._class_name}.form_valid_message is not set. Define "
                f"{self._class_name}.form_valid_message, or override "
                f"{self._class_name}.get_form_valid_message()."
            )

        if not isinstance(self.form_valid_message, (str, Promise)):
            raise ImproperlyConfigured(
                f"{self._class_name}.form_valid_message must be a str or Promise."
            )

        return force_str(self.form_valid_message)

    def form_valid(self, form):
        """
        Set the "form valid" message for standard form validation
        """
        response = super().form_valid(form)
        self.messages.success(
            self.get_form_valid_message(), fail_silently=True
        )
        return response

    def delete(self, *args, **kwargs):
        """
        Set the "form valid" message for delete form validation
        """
        response = super().delete(*args, **kwargs)
        self.messages.success(
            self.get_form_valid_message(), fail_silently=True
        )
        return response


class FormInvalidMessageMixin(MessageMixin):
    """
    Set a string to be sent via Django's messages framework when a form
    fails validation.
    """

    form_invalid_message = None

    def get_form_invalid_message(self):
        """
        Validate that form_invalid_message is set correctly.
        """
        if self.form_invalid_message is None:
            raise ImproperlyConfigured(
                f"{self._class_name}.form_invalid_message is not set. Define "
                f"{self._class_name}.form_invalid_message, or override "
                f"{self._class_name}.get_form_invalid_message()."
            )

        if not isinstance(self.form_invalid_message, (str, Promise)):
            raise ImproperlyConfigured(
                f"{self._class_name}.form_invalid_message must be a str or Promise."
            )

        return force_str(self.form_invalid_message)

    def form_invalid(self, form):
        """
        Set the "form invalid" message for standard form validation
        """
        response = super().form_invalid(form)
        self.messages.error(
            self.get_form_invalid_message(), fail_silently=True
        )
        return response


class FormMessagesMixin(FormValidMessageMixin, FormInvalidMessageMixin):
    """
    Set messages to be sent whether a form is valid or invalid.
    """



"""
django-braces mixins library
----------------------------

Mixins to make Django's generic class-based views simpler.

:copyright: (c) 2013 by Kenneth Love and Chris Jones
:license: BSD 3-clause. See LICENSE for more details
"""

__title__ = "braces"
__version__ = "1.17.0"
__author__ = "Kenneth Love and Chris Jones"
__license__ = "BSD 3-clause"
__copyright__ = "Copyright 2013 Kenneth Love and Chris Jones"



import os
from django.conf import settings
from tests import settings as test_settings


def pytest_configure():
    """Setup Django settings"""
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")
    settings.configure(default_settings=test_settings)



"""Testing backend for object level permissions"""
from tests.helpers import PermissionChecker


class PermissionsCheckerBackend:
    """
    Custom Permission Backend for testing Object Level Permissions.
    """
    supports_object_permissions = True
    supports_anonymous_user = True
    supports_inactive_user = True

    @staticmethod
    def authenticate():
        """Required for a backend"""
        return None

    @staticmethod
    def has_perm(user_obj, perm, obj=None):
        """Used for checking permissions using the `PermissionChecker`"""
        check = PermissionChecker(user_obj)
        return check.has_perm(perm, obj)

    @staticmethod
    def has_perms(user_obj, perms: list[str], obj=None):
        """Used for checking multiple permissions using the `PermissionChecker`"""
        check = PermissionChecker(user_obj)
        return check.has_perms(perms, obj)



from django.conf.global_settings import *

DEBUG = False
TEMPLATE_DEBUG = DEBUG

TIME_ZONE = "UTC"
LANGUAGE_CODE = "en-US"
SITE_ID = 1
USE_TZ = True

SECRET_KEY = "local"

ROOT_URLCONF = "tests.urls"

DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}
}

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
    'tests.backends.PermissionsCheckerBackend',
)

MIDDLEWARE_CLASSES = [
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

MIDDLEWARE = MIDDLEWARE_CLASSES

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
)

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "tests",
)

PASSWORD_HASHERS = ("django.contrib.auth.hashers.MD5PasswordHasher",)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]
AUTH_PASSWORD_VALIDATORS = []



class UserKwargModelFormMixin:
    """
    Generic model form mixin for popping user out of the kwargs and
    attaching it to the instance.

    This mixin must precede forms.ModelForm/forms.Form. The form is not
    expecting these kwargs to be passed in, so they must be popped off before
    anything else is done.
    """

    def __init__(self, *args, **kwargs):
        """Remove the user from **kwargs and assign it on the object"""
        self.user = kwargs.pop("user", None)
        super().__init__(*args, **kwargs)



from django import forms

from braces.forms import UserKwargModelFormMixin

from .models import Article


class FormWithUserKwarg(UserKwargModelFormMixin, forms.Form):
    """This form will get a `user` kwarg"""
    field1 = forms.CharField()


class ArticleForm(forms.ModelForm):
    """This form represents an Article"""
    class Meta:
        model = Article
        fields = ["author", "title", "body", "slug"]



from ._access import (
    AnonymousRequiredMixin,
    GroupRequiredMixin,
    LoginRequiredMixin,
    MultiplePermissionsRequiredMixin,
    PermissionRequiredMixin,
    StaffuserRequiredMixin,
    SuperuserRequiredMixin,
    UserPassesTestMixin,
    SSLRequiredMixin,
    RecentLoginRequiredMixin,
)
from ._ajax import (
    AjaxResponseMixin,
    JSONRequestResponseMixin,
    JSONResponseMixin,
    JsonRequestResponseMixin,
)
from ._forms import (
    CsrfExemptMixin,
    FormInvalidMessageMixin,
    FormMessagesMixin,
    FormValidMessageMixin,
    MessageMixin,
    SuccessURLRedirectListMixin,
    UserFormKwargsMixin,
)
from ._other import (
    AllVerbsMixin,
    CanonicalSlugDetailMixin,
    SetHeadlineMixin,
    StaticContextMixin,
    HeaderMixin,
    CacheControlMixin,
    NeverCacheMixin
)
from ._queries import (
    OrderableListMixin,
    PrefetchRelatedMixin,
    SelectRelatedMixin,
)

__all__ = [
    "AjaxResponseMixin",
    "AllVerbsMixin",
    "AnonymousRequiredMixin",
    "CacheControlMixin",
    "CanonicalSlugDetailMixin",
    "CsrfExemptMixin",
    "FormInvalidMessageMixin",
    "FormMessagesMixin",
    "FormValidMessageMixin",
    "GroupRequiredMixin",
    "HeaderMixin",
    "JSONRequestResponseMixin",
    "JsonRequestResponseMixin",
    "JSONResponseMixin",
    "LoginRequiredMixin",
    "MessageMixin",
    "MultiplePermissionsRequiredMixin",
    "NeverCacheMixin",
    "OrderableListMixin",
    "PermissionRequiredMixin",
    "PrefetchRelatedMixin",
    "SelectRelatedMixin",
    "SetHeadlineMixin",
    "StaffuserRequiredMixin",
    "StaticContextMixin",
    "SuccessURLRedirectListMixin",
    "SuperuserRequiredMixin",
    "UserFormKwargsMixin",
    "UserPassesTestMixin",
    "SSLRequiredMixin",
    "RecentLoginRequiredMixin",
]



from django.contrib.auth.views import LoginView
from django.urls import path
from django.urls import include, re_path
from . import views

urlpatterns = [
    # LoginRequiredMixin tests
    path("login_required/", views.LoginRequiredView.as_view()),
    # AnonymousRequiredView tests
    path(
        "unauthenticated_view/",
        views.AnonymousRequiredView.as_view(),
        name="unauthenticated_view",
    ),
    path(
        "authenticated_view/",
        views.AuthenticatedView.as_view(),
        name="authenticated_view",
    ),
    # AjaxResponseMixin tests
    path("ajax_response/", views.AjaxResponseView.as_view()),
    # CreateAndRedirectToEditView tests
    path("article/create/", views.CreateArticleView.as_view()),
    path(
        "article/<int:pk>/edit/",
        views.EditArticleView.as_view(),
        name="edit_article",
    ),
    path(
        "article_list/create/",
        views.CreateArticleAndRedirectToListView.as_view(),
    ),
    path(
        "article_list_bad/create/",
        views.CreateArticleAndRedirectToListViewBad.as_view(),
    ),
    path(
        "article_list/",
        views.ArticleListView.as_view(),
        name="article_list",
    ),
    # CanonicalSlugDetailMixin tests
    re_path(
        r"^article-canonical/(?P<pk>\d+)-(?P<slug>[-\w]+)/$",
        views.CanonicalSlugDetailView.as_view(),
        name="canonical_slug",
    ),
    re_path(
        r"^article-canonical-override/(?P<pk>\d+)-(?P<slug>[-\w]+)/$",
        views.OverriddenCanonicalSlugDetailView.as_view(),
        name="canonical_override",
    ),
    re_path(
        r"^article-canonical-custom-kwargs/(?P<my_pk>\d+)-(?P<my_slug>[-\w]+)/$",
        views.CanonicalSlugDetailCustomUrlKwargsView.as_view(),
        name="canonical_custom_kwargs",
    ),
    re_path(
        r"^article-canonical-model/(?P<pk>\d+)-(?P<slug>[-\w]+)/$",
        views.ModelCanonicalSlugDetailView.as_view(),
        name="canonical_model",
    ),
    # UserFormKwargsMixin tests
    path("form_with_user_kwarg/", views.FormWithUserKwargView.as_view()),
    # SetHeadlineMixin tests
    path("headline/", views.HeadlineView.as_view(), name="headline"),
    path("headline/lazy/", views.LazyHeadlineView.as_view()),
    re_path(r"^headline/(?P<s>[\w-]+)/$", views.DynamicHeadlineView.as_view()),
    # ExtraContextMixin tests
    path("context/", views.ContextView.as_view(), name="context"),
    # PermissionRequiredMixin tests
    path("permission_required/", views.PermissionRequiredView.as_view()),
    path("object_level_permission_required/", views.PermissionRequiredView.as_view(object_level_permissions=True)),
    # MultiplePermissionsRequiredMixin tests
    path(
        "multiple_permissions_required/",
        views.MultiplePermissionsRequiredView.as_view(),
    ),
    path(
        "multiple_object_level_permissions_required/",
        views.MultiplePermissionsRequiredView.as_view(object_level_permissions=True),
    ),
    # SuperuserRequiredMixin tests
    path("superuser_required/", views.SuperuserRequiredView.as_view()),
    # StaffuserRequiredMixin tests
    path("staffuser_required/", views.StaffuserRequiredView.as_view()),
    # GroupRequiredMixin tests
    path("group_required/", views.GroupRequiredView.as_view()),
    # UserPassesTestMixin tests
    path("user_passes_test/", views.UserPassesTestView.as_view()),
    # UserPassesTestMixin tests
    path(
        "user_passes_test_not_implemented/",
        views.UserPassesTestNotImplementedView.as_view(),
    ),
    # CsrfExemptMixin tests
    path("csrf_exempt/", views.CsrfExemptView.as_view()),
    # JSONResponseMixin tests
    path("simple_json/", views.SimpleJsonView.as_view()),
    path(
        "simple_json_custom_encoder/", views.CustomJsonEncoderView.as_view()
    ),
    path("simple_json_400/", views.SimpleJsonBadRequestView.as_view()),
    path("article_list_json/", views.ArticleListJsonView.as_view()),
    # JsonRequestResponseMixin tests
    path("json_request/", views.JsonRequestResponseView.as_view()),
    path("json_bad_request/", views.JsonBadRequestView.as_view()),
    path(
        "json_custom_bad_request/", views.JsonCustomBadRequestView.as_view()
    ),
    # FormMessagesMixin tests
    path("form_messages/", views.FormMessagesView.as_view()),
    # AllVerbsMixin tests
    path("all_verbs/", views.AllVerbsView.as_view()),
    path(
        "all_verbs_no_handler/", views.AllVerbsView.as_view(all_handler=None)
    ),
    # SSLRequiredMixin tests
    path("sslrequired/", views.SSLRequiredView.as_view()),
    # RecentLoginRequiredMixin tests
    path("recent_login/", views.RecentLoginRequiredView.as_view()),
    path("outdated_login/", views.RecentLoginRequiredView.as_view()),
    # HeaderMixin tests
    path('headers/attribute/', views.AttributeHeaderView.as_view()),
    path('headers/method/', views.MethodHeaderView.as_view()),
    path('headers/existing/', views.ExistingHeaderView.as_view()),
    # CacheControlMixin tests
    path('cachecontrol/public/', views.CacheControlPublicView.as_view()),
    # NeverCacheMixin tests
    path('nevercache/', views.NeverCacheView.as_view()),
]

urlpatterns += [
    path(
        "accounts/login/", LoginView.as_view(template_name="blank.html")
    ),
    path("auth/login/", LoginView.as_view(template_name="blank.html")),
]

urlpatterns += [
    path(
        "article-canonical-namespaced/",
        include(
            ("tests.urls_namespaced", "tests"), namespace="some_namespace"
        ),
    ),
]



import os
import re
from setuptools import setup

NAME = "braces"


def _add_default(m):
    """Add on a default"""
    attr_name, attr_value = m.groups()
    return ((attr_name, attr_value.strip("\"'")),)


def parse_dist_meta():
    """Extract metadata information from ``$dist/__init__.py``."""

    re_meta = re.compile(r"__(\w+?)__\s*=\s*(.*)")
    re_doc = re.compile(r'^"""(.+?)"""')
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, NAME, "__init__.py")) as meta_fh:
        distmeta = {}
        for line in meta_fh:
            if line.strip() == "# -eof meta-":
                break
            match = re_meta.match(line.strip())
            if match:
                distmeta.update(_add_default(match))
        return distmeta


meta = parse_dist_meta()

setup(
    name="django-braces",
    version=meta["version"],
    description="Reusable, generic mixins for Django",
    long_description="Mixins to add easy functionality to Django class-based views, forms, and models.",
    keywords="django, views, forms, mixins",
    author="Kenneth Love <kenneth@brack3t.com>, Chris Jones <chris@brack3t.com>",
    author_email="devs@brack3t.com",
    url="https://github.com/brack3t/django-braces/",
    license="BSD",
    packages=["braces", "braces.views"],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Development Status :: 6 - Mature",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=["Django>=2.2"],
)



from django.urls import re_path

from . import views


urlpatterns = [
    re_path(
        r"^article/(?P<pk>\d+)-(?P<slug>[\w-]+)/$",
        views.CanonicalSlugDetailView.as_view(),
        name="namespaced_article",
    ),
]



import codecs

from django.contrib.auth.models import User
from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _
from django.views.generic import (
    View,
    UpdateView,
    FormView,
    TemplateView,
    ListView,
    DetailView,
    CreateView,
)

from braces import views

from .models import Article, CanonicalArticle
from .forms import ArticleForm, FormWithUserKwarg
from .helpers import SetJSONEncoder


class OkView(View):
    """
    A view which simply returns "OK" for every request.
    """

    def get(self, request):
        """Everything is going to be OK"""
        return HttpResponse("OK")

    def post(self, request):
        """Get it?"""
        return self.get(request)

    def put(self, request):
        """Get it?"""
        return self.get(request)

    def delete(self, request):
        """Get it?"""
        return self.get(request)


class LoginRequiredView(views.LoginRequiredMixin, OkView):
    """
    A view for testing LoginRequiredMixin.
    """


class AnonymousRequiredView(views.AnonymousRequiredMixin, OkView):
    """
    A view for testing AnonymousRequiredMixin. Should accept
    unauthenticated users and redirect authenticated users to the
    authenticated_redirect_url set on the view.
    """

    authenticated_redirect_url = "/authenticated_view/"


class AuthenticatedView(views.LoginRequiredMixin, OkView):
    """
    A view for testing AnonymousRequiredMixin. Should accept
    authenticated users.
    """


class AjaxResponseView(views.AjaxResponseMixin, OkView):
    """
    A view for testing AjaxResponseMixin.
    """

    def get_ajax(self, request):
        """Everything will eventually be OK"""
        return HttpResponse("AJAX_OK")

    def post_ajax(self, request):
        """Get it?"""
        return self.get_ajax(request)

    def put_ajax(self, request):
        """Get it?"""
        return self.get_ajax(request)

    def delete_ajax(self, request):
        """Get it?"""
        return self.get_ajax(request)


class SimpleJsonView(views.JSONResponseMixin, View):
    """
    A view for testing JSONResponseMixin's render_json_response() method.
    """

    def get(self, request):
        """Send back some JSON"""
        object = {"username": request.user.username}
        return self.render_json_response(object)


class CustomJsonEncoderView(views.JSONResponseMixin, View):
    """
    A view for testing JSONResponseMixin's `json_encoder_class` attribute
    with custom JSONEncoder class.
    """

    json_encoder_class = SetJSONEncoder

    def get(self, request):
        """Send back some JSON"""
        object = {"numbers": set([1, 2, 3])}
        return self.render_json_response(object)


class SimpleJsonBadRequestView(views.JSONResponseMixin, View):
    """
    A view for testing JSONResponseMixin's render_json_response() method with
    400 HTTP status code.
    """

    def get(self, request):
        """Send back some JSON"""
        object = {"username": request.user.username}
        return self.render_json_response(object, status=400)


class ArticleListJsonView(views.JSONResponseMixin, View):
    """
    A view for testing JSONResponseMixin's render_json_object_response()
    method.
    """

    def get(self, request):
        """Send back some JSON"""
        queryset = Article.objects.all()
        return self.render_json_object_response(queryset, fields=("title",))


class JsonRequestResponseView(views.JsonRequestResponseMixin, View):
    """
    A view for testing JsonRequestResponseMixin's json conversion
    """

    def post(self, request):
        """Send back some JSON"""
        return self.render_json_response(self.request_json)


class JsonBadRequestView(views.JsonRequestResponseMixin, View):
    """
    A view for testing JsonRequestResponseMixin's require_json
    and render_bad_request_response methods
    """

    require_json = True

    def post(self, request, *args, **kwargs):
        """Send back some JSON"""
        return self.render_json_response(self.request_json)


class JsonCustomBadRequestView(views.JsonRequestResponseMixin, View):
    """
    A view for testing JsonRequestResponseMixin's
    render_bad_request_response method with a custom error message
    """

    def post(self, request, *args, **kwargs):
        """Handle the POST request"""
        if not self.request_json:
            return self.render_bad_request_response({"error": "you messed up"})
        return self.render_json_response(self.request_json)


class CreateArticleView(CreateView):
    """
    View for testing CreateAndRedirectEditToView.
    """

    fields = ["author", "title", "body", "slug"]
    model = Article
    template_name = "form.html"


class EditArticleView(UpdateView):
    """
    View for testing CreateAndRedirectEditToView.
    """

    model = Article
    template_name = "form.html"


class CreateArticleAndRedirectToListView(
    views.SuccessURLRedirectListMixin, CreateArticleView
):
    """
    View for testing SuccessURLRedirectListMixin
    """

    success_list_url = "article_list"


class CreateArticleAndRedirectToListViewBad(
    views.SuccessURLRedirectListMixin, CreateArticleView
):
    """
    View for testing SuccessURLRedirectListMixin
    """

    success_list_url = None


class ArticleListView(views.SelectRelatedMixin, ListView):
    """
    A list view for articles, required for testing SuccessURLRedirectListMixin.

    Also used to test SelectRelatedMixin.
    """

    model = Article
    template_name = "blank.html"
    select_related = ("author",)


class ArticleListViewWithCustomQueryset(views.SelectRelatedMixin, ListView):
    """
    Another list view for articles, required to test SelectRelatedMixin.
    """

    queryset = Article.objects.select_related("author").prefetch_related(
        "article_set"
    )
    template_name = "blank.html"
    select_related = ()


class FormWithUserKwargView(views.UserFormKwargsMixin, FormView):
    """
    View for testing UserFormKwargsMixin.
    """

    form_class = FormWithUserKwarg
    template_name = "form.html"

    def form_valid(self, form):
        """A simple response to watch for"""
        return HttpResponse(f"username: {form.user.username}")


class HeadlineView(views.SetHeadlineMixin, TemplateView):
    """
    View for testing SetHeadlineMixin.
    """

    template_name = "blank.html"
    headline = "Test headline"


class LazyHeadlineView(views.SetHeadlineMixin, TemplateView):
    """
    View for testing SetHeadlineMixin.
    """

    template_name = "blank.html"
    headline = _("Test Headline")


class ContextView(views.StaticContextMixin, TemplateView):
    """View for testing StaticContextMixin."""

    template_name = "blank.html"
    static_context = {"test": True}


class DynamicHeadlineView(views.SetHeadlineMixin, TemplateView):
    """
    View for testing SetHeadlineMixin's get_headline() method.
    """

    template_name = "blank.html"

    def get_headline(self):
        """Return the headline passed in via kwargs"""
        return self.kwargs["s"]


class PermissionRequiredView(views.PermissionRequiredMixin, OkView):
    """
    View for testing PermissionRequiredMixin.
    """

    permission_required = "auth.add_user"

    def get_object(self):
        """
        Get the article object for the query parameter's primary key
        """
        article_obj = None
        object_pk = self.request.GET.get("pk")
        if object_pk:
            article_obj = Article.objects.get(pk=object_pk)
        return article_obj


class MultiplePermissionsRequiredView(
    views.MultiplePermissionsRequiredMixin, OkView
):
    permissions = {
        "all": ["tests.add_article", "tests.change_article"],
        "any": ["auth.add_user", "auth.change_user"],
    }

    def get_object(self):
        """
        Get the article object for the query parameter's primary key
        """
        article_obj = None
        object_pk = self.request.GET.get("pk")
        if object_pk:
            article_obj = Article.objects.get(pk=object_pk)
        return article_obj




class SuperuserRequiredView(views.SuperuserRequiredMixin, OkView):
    """Require a superuser"""


class StaffuserRequiredView(views.StaffuserRequiredMixin, OkView):
    """Require a user marked as `is_staff`"""


class CsrfExemptView(views.CsrfExemptMixin, OkView):
    """Ignore CSRF"""


class AuthorDetailView(views.PrefetchRelatedMixin, ListView):
    """A basic detail view to test prefetching"""
    model = User
    prefetch_related = ["article_set"]
    template_name = "blank.html"


class OrderableListView(views.OrderableListMixin, ListView):
    """A basic list view to test ordering the output"""
    model = Article
    orderable_columns = (
        "id",
        "title",
    )
    orderable_columns_default = "id"


class CanonicalSlugDetailView(views.CanonicalSlugDetailMixin, DetailView):
    """A basic detail view to test a canonical slug"""
    model = Article
    template_name = "blank.html"


class OverriddenCanonicalSlugDetailView(views.CanonicalSlugDetailMixin, DetailView):
    """A basic detail view to test an overridden slug"""
    model = Article
    template_name = "blank.html"

    def get_canonical_slug(self):
        """Give back a different, encoded slug. My slug secrets are safe"""
        return codecs.encode(self.get_object().slug, "rot_13")


class CanonicalSlugDetailCustomUrlKwargsView(views.CanonicalSlugDetailMixin, DetailView):
    """A basic detail view to test a slug with custom URL stuff"""
    model = Article
    template_name = "blank.html"
    pk_url_kwarg = "my_pk"
    slug_url_kwarg = "my_slug"


class ModelCanonicalSlugDetailView(views.CanonicalSlugDetailMixin, DetailView):
    """A basic detail view to test a model with a canonical slug"""
    model = CanonicalArticle
    template_name = "blank.html"


class FormMessagesView(views.FormMessagesMixin, CreateView):
    """A basic form view to test valid/invalid messages"""
    form_class = ArticleForm
    form_invalid_message = _("Invalid")
    form_valid_message = _("Valid")
    model = Article
    success_url = "/form_messages/"
    template_name = "form.html"


class GroupRequiredView(views.GroupRequiredMixin, OkView):
    """Is everything OK in this group?"""
    group_required = "test_group"


class UserPassesTestView(views.UserPassesTestMixin, OkView):
    """Did I pass a test?"""
    def test_func(self, user):
        return (
            user.is_staff
            and not user.is_superuser
            and user.email.endswith("@mydomain.com")
        )


class UserPassesTestLoginRequiredView(
    views.LoginRequiredMixin, views.UserPassesTestMixin, OkView
):
    """Am I logged in _and_ passing a test?"""
    def test_func(self, user):
        return (
            user.is_staff
            and not user.is_superuser
            and user.email.endswith("@mydomain.com")
        )


class UserPassesTestNotImplementedView(views.UserPassesTestMixin, OkView):
    """The test went missing?"""
    pass


class AllVerbsView(views.AllVerbsMixin, View):
    """I know, like, all the verbs"""
    def all(self, request, *args, **kwargs):
        return HttpResponse("All verbs return this!")


class SSLRequiredView(views.SSLRequiredMixin, OkView):
    """Speak friend and enter"""
    pass


class RecentLoginRequiredView(views.RecentLoginRequiredMixin, OkView):
    """
    A view for testing RecentLoginRequiredMixin.
    """


class AttributeHeaderView(views.HeaderMixin, OkView):
    """Set headers in an attribute w/o a template render class"""
    headers = {
        "X-DJANGO-BRACES-1": 1,
        "X-DJANGO-BRACES-2": 2,
    }


class MethodHeaderView(views.HeaderMixin, OkView):
    """Set headers in a method w/o a template render class"""
    def get_headers(self, request):
        return {
            "X-DJANGO-BRACES-1": 1,
            "X-DJANGO-BRACES-2": 2,
        }


class AuxiliaryHeaderView(View):
    """A view with a header already set"""
    def dispatch(self, request, *args, **kwargs):
        response = HttpResponse("OK with headers")
        response["X-DJANGO-BRACES-EXISTING"] = "value"
        return response


class ExistingHeaderView(views.HeaderMixin, AuxiliaryHeaderView):
    """A view trying to override a parent's header"""
    headers = {
        'X-DJANGO-BRACES-EXISTING': 'other value'
    }


class CacheControlPublicView(views.CacheControlMixin, OkView):
    """A public-cached page with a 60 second timeout"""
    cachecontrol_public = True
    cachecontrol_max_age = 60


class NeverCacheView(views.NeverCacheMixin, OkView):
    """
    View that will never be cached upstream.
    """



import datetime

import pytest

from django import test
from django.contrib.auth.models import Permission
from django.test.utils import override_settings
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import Http404, HttpResponse
from django.utils.encoding import force_str
from django.utils.timezone import make_aware, get_current_timezone

from django.urls import reverse_lazy

from .factories import (
    GroupFactory,
    UserFactory,
    UserObjectPermissionsFactory,
    ArticleFactory,
)
from .helpers import TestViewHelper
from .views import (
    PermissionRequiredView,
    MultiplePermissionsRequiredView,
    SuperuserRequiredView,
    StaffuserRequiredView,
    LoginRequiredView,
    GroupRequiredView,
    UserPassesTestView,
    UserPassesTestNotImplementedView,
    AnonymousRequiredView,
    SSLRequiredView,
    RecentLoginRequiredView,
    UserPassesTestLoginRequiredView,
)


@pytest.mark.django_db
class _TestAccessBasicsMixin(TestViewHelper):
    """
    A set of basic tests for access mixins.
    """

    view_url = None

    def build_authorized_user(self):
        """
        Returns user authorized to access view.
        """
        raise NotImplementedError

    def build_unauthorized_user(self):
        """
        Returns user not authorized to access view.
        """
        raise NotImplementedError

    def test_success(self):
        """
        If user is authorized then view should return normal response.
        """
        user = self.build_authorized_user()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))

    def test_redirects_to_login(self):
        """
        Browser should be redirected to login page if user is not authorized
        to view this page.
        """
        user = self.build_unauthorized_user()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertRedirects(resp, "/accounts/login/?next={0}".format(self.view_url))

    def test_raise_permission_denied(self):
        """
        PermissionDenied should be raised if user is not authorized and
        raise_exception attribute is set to True.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)

        with self.assertRaises(PermissionDenied):
            self.dispatch_view(req, raise_exception=True)

    def test_raise_custom_exception(self):
        """
        Http404 should be raised if user is not authorized and
        raise_exception attribute is set to Http404.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)

        with self.assertRaises(Http404):
            self.dispatch_view(req, raise_exception=Http404)

    def test_raise_func_pass(self):
        """
        An exception should be raised if user is not authorized and
        raise_exception attribute is set to a function that returns nothing.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)

        def func(request):
            pass

        with self.assertRaises(PermissionDenied):
            self.dispatch_view(req, raise_exception=func)

    def test_raise_func_response(self):
        """
        A custom response should be returned if user is not authorized and
        raise_exception attribute is set to a function that returns a response.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)

        def func(request):
            return HttpResponse("CUSTOM")

        resp = self.dispatch_view(req, raise_exception=func)
        assert resp.status_code == 200
        assert force_str(resp.content) == "CUSTOM"

    def test_raise_func_false(self):
        """
        PermissionDenied should be raised, if a custom raise_exception
        function does not return HttpResponse or StreamingHttpResponse.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)

        def func(request):
            return False

        with self.assertRaises(PermissionDenied):
            self.dispatch_view(req, raise_exception=func)

    def test_raise_func_raises(self):
        """
        A custom exception should be raised if user is not authorized and
        raise_exception attribute is set to a callable that raises an
        exception.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)

        def func(request):
            raise Http404

        with self.assertRaises(Http404):
            self.dispatch_view(req, raise_exception=func)

    def test_custom_login_url(self):
        """
        Login url should be customizable.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)
        resp = self.dispatch_view(req, login_url="/login/")
        self.assertEqual("/login/?next={0}".format(self.view_url), resp["Location"])

        # Test with reverse_lazy
        resp = self.dispatch_view(req, login_url=reverse_lazy("headline"))
        self.assertEqual("/headline/?next={0}".format(self.view_url), resp["Location"])

    def test_custom_redirect_field_name(self):
        """
        Redirect field name should be customizable.
        """
        user = self.build_unauthorized_user()
        req = self.build_request(user=user, path=self.view_url)
        resp = self.dispatch_view(req, redirect_field_name="foo")
        expected_url = "/accounts/login/?foo={0}".format(self.view_url)
        self.assertEqual(expected_url, resp["Location"])

    @override_settings(LOGIN_URL=None)
    def test_get_login_url_raises_exception(self):
        """
        Test that get_login_url from AccessMixin raises
        ImproperlyConfigured.
        """
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(self.build_request(path=self.view_url), login_url=None)

    def test_get_redirect_field_name_raises_exception(self):
        """
        Test that get_redirect_field_name from AccessMixin raises
        ImproperlyConfigured.
        """
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(
                self.build_request(path=self.view_url),
                redirect_field_name=None,
            )

    @override_settings(LOGIN_URL="/auth/login/")
    def test_overridden_login_url(self):
        """
        Test that login_url is not set in stone on module load but can be
        overridden dynamically.
        """
        user = self.build_unauthorized_user()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertRedirects(resp, "/auth/login/?next={0}".format(self.view_url))

    def test_redirect_unauthenticated(self):
        resp = self.dispatch_view(
            self.build_request(path=self.view_url),
            raise_exception=True,
            redirect_unauthenticated_users=True,
        )
        assert resp.status_code == 302
        assert resp["Location"] == "/accounts/login/?next={0}".format(self.view_url)

    def test_redirect_unauthenticated_false(self):
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(path=self.view_url),
                raise_exception=True,
                redirect_unauthenticated_users=False,
            )


@pytest.mark.django_db
class TestLoginRequiredMixin(TestViewHelper, test.TestCase):
    """Scenarios around requiring an authenticated session"""

    view_class = LoginRequiredView
    view_url = "/login_required/"

    def test_anonymous(self):
        """Anonymous users should be redirected"""
        resp = self.client.get(self.view_url)
        self.assertRedirects(resp, "/accounts/login/?next=/login_required/")

    def test_anonymous_raises_exception(self):
        """Anonymous users should raise an exception"""
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(path=self.view_url), raise_exception=True
            )

    def test_authenticated(self):
        """Authenticated users should get 'OK'"""
        user = UserFactory()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        assert resp.status_code == 200
        assert force_str(resp.content) == "OK"

    def test_anonymous_redirects(self):
        """Anonymous users are redirected with a 302"""
        resp = self.dispatch_view(
            self.build_request(path=self.view_url),
            raise_exception=True,
            redirect_unauthenticated_users=True,
        )
        assert resp.status_code == 302
        assert resp["Location"] == "/accounts/login/?next=/login_required/"


@pytest.mark.django_db
class TestChainedLoginRequiredMixin(TestViewHelper, test.TestCase):
    """
    Tests for LoginRequiredMixin combined with another AccessMixin.
    """

    view_class = UserPassesTestLoginRequiredView
    view_url = "/chained_view/"

    def assert_redirect_to_login(self, response):
        """
        Check that the response is a redirect to the login view.
        """
        assert response.status_code == 302
        assert response["Location"] == "/accounts/login/?next=/chained_view/"

    def test_anonymous(self):
        """
        Check that anonymous users redirect to login by default.
        """
        resp = self.dispatch_view(self.build_request(path=self.view_url))
        self.assert_redirect_to_login(resp)

    def test_anonymous_raises_exception(self):
        """
        Check that when anonymous users hit a view that has only
        raise_exception set, they get a PermissionDenied.
        """
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(path=self.view_url), raise_exception=True
            )

    def test_authenticated_raises_exception(self):
        """
        Check that when authenticated users hit a view that has raise_exception
        set, they get a PermissionDenied.
        """
        user = UserFactory()
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(path=self.view_url, user=user),
                raise_exception=True,
            )
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(path=self.view_url, user=user),
                raise_exception=True,
                redirect_unauthenticated_users=True,
            )

    def test_anonymous_redirects(self):
        """
        Check that anonymous users are redirected to login when raise_exception
        is overridden by redirect_unauthenticated_users.
        """
        resp = self.dispatch_view(
            self.build_request(path=self.view_url),
            raise_exception=True,
            redirect_unauthenticated_users=True,
        )
        self.assert_redirect_to_login(resp)


@pytest.mark.django_db
class TestAnonymousRequiredMixin(TestViewHelper, test.TestCase):
    """
    Tests for AnonymousRequiredMixin.
    """

    view_class = AnonymousRequiredView
    view_url = "/unauthenticated_view/"

    def test_anonymous(self):
        """
        As a non-authenticated user, it should be possible to access
        the URL.
        """
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))

        # Test with reverse_lazy
        resp = self.dispatch_view(
            self.build_request(), login_url=reverse_lazy(self.view_url)
        )
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))

    def test_authenticated(self):
        """
        Check that the authenticated user has been successfully directed
        to the appropriate view.
        """
        user = UserFactory()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertEqual(302, resp.status_code)

        resp = self.client.get(self.view_url, follow=True)
        self.assertRedirects(resp, "/authenticated_view/")

    def test_no_url(self):
        """View should raise an exception if no URL is provided"""
        self.view_class.authenticated_redirect_url = None
        user = UserFactory()
        self.client.login(username=user.username, password="asdf1234")
        with self.assertRaises(ImproperlyConfigured):
            self.client.get(self.view_url)

    def test_bad_url(self):
        """Redirection can be misconfigured"""
        self.view_class.authenticated_redirect_url = "/epicfailurl/"
        user = UserFactory()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url, follow=True)
        self.assertEqual(404, resp.status_code)


@pytest.mark.django_db
class TestPermissionRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
    """Scenarios around requiring a permission"""

    view_class = PermissionRequiredView
    view_url = "/permission_required/"

    def build_authorized_user(self):
        """Create a user with permissions"""
        return UserFactory(permissions=["auth.add_user"])

    def build_unauthorized_user(self):
        """Create a user without permissions"""
        return UserFactory()

    def test_invalid_permission(self):
        """
        ImproperlyConfigured exception should be raised in two situations:
        if permission is None or if permission has invalid name.
        """
        with self.assertRaises(ImproperlyConfigured):
            self.dispatch_view(self.build_request(), permission_required=None)

    def test_object_level_permissions(self):
        """
        Tests that object level permissions perform as expected, where object level permissions and
        global level permissions
        """
        # Arrange
        article = ArticleFactory()
        self.view_class = PermissionRequiredView
        self.view_url = f"/object_level_permission_required/?pk={article.pk}"
        tests_add_article = Permission.objects.get(codename="add_article")
        permissions = "tests.add_article"
        valid_user = UserFactory(permissions=[permissions])
        invalid_user_1 = UserFactory(permissions=["auth.add_user"])
        invalid_user_2 = UserFactory(permissions=[permissions])
        UserObjectPermissionsFactory(
            user=valid_user, permission=tests_add_article, article_object=article
        )
        # Act
        valid_req = self.build_request(path=self.view_url, user=valid_user)
        valid_resp = self.dispatch_view(
            valid_req,
            permission_required=permissions,
            object_level_permissions=True,
            raise_exception=True,
        )
        invalid_req_1 = self.build_request(path=self.view_url, user=invalid_user_1)
        invalid_req_2 = self.build_request(path=self.view_url, user=invalid_user_2)
        # Assert
        self.assertEqual(valid_resp.status_code, 200)
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                invalid_req_1,
                permission_required=permissions,
                object_level_permissions=True,
                raise_exception=True,
            )
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                invalid_req_2,
                permission_required=permissions,
                object_level_permissions=True,
                raise_exception=True,
            )


@pytest.mark.django_db
class TestMultiplePermissionsRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
    """Scenarios around requiring multiple permissions"""

    view_class = MultiplePermissionsRequiredView
    view_url = "/multiple_permissions_required/"

    def build_authorized_user(self):
        """Get a user with permissions"""
        return UserFactory(
            permissions=[
                "tests.add_article",
                "tests.change_article",
                "auth.change_user",
            ]
        )

    def build_unauthorized_user(self):
        """Get a user without the important permissions"""
        return UserFactory(permissions=["tests.add_article"])

    def test_redirects_to_login(self):
        """
        User should be redirected to login page if he or she does not have
        sufficient permissions.
        """
        url = "/multiple_permissions_required/"
        test_cases = (
            # missing one permission from 'any'
            ["tests.add_article", "tests.change_article"],
            # missing one permission from 'all'
            ["tests.add_article", "auth.add_user"],
            # no permissions at all
            [],
        )

        for permissions in test_cases:
            user = UserFactory(permissions=permissions)
            self.client.login(username=user.username, password="asdf1234")
            resp = self.client.get(url)
            self.assertRedirects(resp, "/accounts/login/?next={0}".format(url))

    def test_invalid_permissions(self):
        """
        ImproperlyConfigured exception should be raised if permissions
        attribute is set incorrectly.
        """
        permissions = (
            None,  # permissions must be set
            (),  # and they must be a dict
            {},  # at least one of 'all', 'any' keys must be present
            {"all": None},  # both all and any must be list or a tuple
            {"all": {"a": 1}},
            {"any": None},
            {"any": {"a": 1}},
        )

        for attr in permissions:
            with self.assertRaises(ImproperlyConfigured):
                self.dispatch_view(self.build_request(), permissions=attr)

    def test_raise_permission_denied(self):
        """
        PermissionDenied should be raised if user does not have sufficient
        permissions and raise_exception is set to True.
        """
        test_cases = (
            # missing one permission from 'any'
            ["tests.add_article", "tests.change_article"],
            # missing one permission from 'all'
            ["tests.add_article", "auth.add_user"],
            # no permissions at all
            [],
        )

        for permissions in test_cases:
            user = UserFactory(permissions=permissions)
            req = self.build_request(user=user)
            with self.assertRaises(PermissionDenied):
                self.dispatch_view(req, raise_exception=True)

    def test_all_permissions_key(self):
        """
        Tests if everything works if only 'all' permissions has been set.
        """
        permissions = {"all": ["auth.add_user", "tests.add_article"]}
        user = UserFactory(permissions=permissions["all"])
        req = self.build_request(user=user)

        resp = self.dispatch_view(req, permissions=permissions)
        self.assertEqual("OK", force_str(resp.content))

        user = UserFactory(permissions=["auth.add_user"])
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(user=user),
                raise_exception=True,
                permissions=permissions,
            )

    def test_any_permissions_key(self):
        """
        Tests if everything works if only 'any' permissions has been set.
        """
        permissions = {"any": ["auth.add_user", "tests.add_article"]}
        user = UserFactory(permissions=["tests.add_article"])
        req = self.build_request(user=user)

        resp = self.dispatch_view(req, permissions=permissions)
        self.assertEqual("OK", force_str(resp.content))

        user = UserFactory(permissions=[])
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(user=user),
                raise_exception=True,
                permissions=permissions,
            )

    def test_all_object_level_permissions_key(self):
        """
        Tests that when a user has all the correct object level permissions, response is OK,
        else forbidden.
        """
        # Arrange
        article = ArticleFactory()
        self.view_class = MultiplePermissionsRequiredView
        self.view_url = f"/multiple_object_level_permissions_required/?pk={article.pk}"
        auth_add_user = Permission.objects.get(codename="add_user")
        tests_add_article = Permission.objects.get(codename="add_article")
        permissions = {"all": ["auth.add_user", "tests.add_article"]}
        valid_user = UserFactory(permissions=permissions["all"])
        invalid_user = UserFactory(permissions=["auth.add_user"])
        UserObjectPermissionsFactory(
            user=valid_user, permission=auth_add_user, article_object=article
        )
        UserObjectPermissionsFactory(
            user=valid_user, permission=tests_add_article, article_object=article
        )
        # Act
        valid_req = self.build_request(path=self.view_url, user=valid_user)
        valid_resp = self.dispatch_view(
            valid_req, permissions=permissions, object_level_permissions=True
        )
        invalid_req = self.build_request(path=self.view_url, user=invalid_user)
        # Arrange
        self.assertEqual(valid_resp.status_code, 200)
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                invalid_req,
                permissions=permissions,
                object_level_permissions=True,
                raise_exception=True,
            )

    def test_any_object_level_permissions_key(self):
        """
        Tests that when a user has any the correct object level permissions, response is OK,
        else forbidden.
        """
        # Arrange
        article = ArticleFactory()
        self.view_url = f"/multiple_object_level_permissions_required/?pk={article.pk}"
        self.view_class = MultiplePermissionsRequiredView
        auth_add_user = Permission.objects.get(codename="add_user")
        tests_add_article = Permission.objects.get(codename="add_article")
        permissions = {"any": ["auth.add_user", "tests.add_article"]}
        user = UserFactory(permissions=[permissions["any"][0]])
        user_1 = UserFactory()
        user_2 = UserFactory(permissions=permissions["any"])
        UserObjectPermissionsFactory(
            user=user, permission=auth_add_user, article_object=article
        )
        UserObjectPermissionsFactory(
            user=user, permission=tests_add_article, article_object=article
        )
        # Act
        valid_req = self.build_request(path=self.view_url, user=user)
        valid_resp = self.dispatch_view(
            valid_req,
            permissions=permissions,
            object_level_permissions=True,
            raise_exception=True,
        )
        invalid_req_1 = self.build_request(path=self.view_url, user=user_1)
        invalid_req_2 = self.build_request(path=self.view_url, user=user_2)
        # Assert
        self.assertEqual(valid_resp.status_code, 200)
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                invalid_req_1,
                permissions=permissions,
                object_level_permissions=True,
                raise_exception=True,
            )
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                invalid_req_2,
                permissions=permissions,
                object_level_permissions=True,
                raise_exception=True,
            )


@pytest.mark.django_db
class TestSuperuserRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
    """Scenarios requiring a superuser"""

    view_class = SuperuserRequiredView
    view_url = "/superuser_required/"

    def build_authorized_user(self):
        """Make a superuser"""
        return UserFactory(is_superuser=True, is_staff=True)

    def build_unauthorized_user(self):
        """Make a non-superuser"""
        return UserFactory()


@pytest.mark.django_db
class TestStaffuserRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
    """Scenarios requiring a staff user"""

    view_class = StaffuserRequiredView
    view_url = "/staffuser_required/"

    def build_authorized_user(self):
        """Hire a user"""
        return UserFactory(is_staff=True)

    def build_unauthorized_user(self):
        """Get a customer"""
        return UserFactory()


@pytest.mark.django_db
class TestGroupRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
    """Scenarios requiring membership in a certain group"""

    view_class = GroupRequiredView
    view_url = "/group_required/"

    def build_authorized_user(self):
        """Get a user with the right group"""
        user = UserFactory()
        group = GroupFactory(name="test_group")
        user.groups.add(group)
        return user

    def build_superuser(self):
        """Get a superuser"""
        user = UserFactory()
        user.is_superuser = True
        user.save()
        return user

    def build_unauthorized_user(self):
        """Just a normal users, not super and no groups"""
        return UserFactory()

    def test_with_string(self):
        """A group name as a string should restrict access"""
        self.assertEqual("test_group", self.view_class.group_required)
        user = self.build_authorized_user()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))

    def test_with_group_list(self):
        """A list of group names should restrict access"""
        group_list = ["test_group", "editors"]
        # the test client will instantiate a new view on request, so we have to
        # modify the class variable (and restore it when the test finished)
        self.view_class.group_required = group_list
        self.assertEqual(group_list, self.view_class.group_required)
        user = self.build_authorized_user()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))
        self.view_class.group_required = "test_group"
        self.assertEqual("test_group", self.view_class.group_required)

    def test_superuser_allowed(self):
        """Superusers should always be allowed, regardless of group rules"""
        user = self.build_superuser()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))

    def test_improperly_configured(self):
        """No group(s) specified should raise ImproperlyConfigured"""
        view = self.view_class()
        view.group_required = None
        with self.assertRaises(ImproperlyConfigured):
            view.get_group_required()

        view.group_required = {"foo": "bar"}
        with self.assertRaises(ImproperlyConfigured):
            view.get_group_required()

    def test_with_unicode(self):
        """Unicode in group names should restrict access"""
        self.view_class.group_required = "niño"
        self.assertEqual("niño", self.view_class.group_required)

        user = self.build_authorized_user()
        group = user.groups.all()[0]
        group.name = "niño"
        group.save()
        self.assertEqual("niño", user.groups.all()[0].name)

        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))
        self.view_class.group_required = "test_group"
        self.assertEqual("test_group", self.view_class.group_required)


@pytest.mark.django_db
class TestUserPassesTestMixin(_TestAccessBasicsMixin, test.TestCase):
    """Scenarios requiring a user to pass a test"""

    view_class = UserPassesTestView
    view_url = "/user_passes_test/"
    view_not_implemented_class = UserPassesTestNotImplementedView
    view_not_implemented_url = "/user_passes_test_not_implemented/"

    # for testing with passing and not passing func_test
    def build_authorized_user(self, is_superuser=False):
        """Get a test-passing user"""
        return UserFactory(
            is_superuser=is_superuser, is_staff=True, email="user@mydomain.com"
        )

    def build_unauthorized_user(self):
        """Get a blank user"""
        return UserFactory()

    def test_with_user_pass(self):
        """Valid username and password should pass the test"""
        user = self.build_authorized_user()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)

        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))

    def test_with_user_not_pass(self):
        """A failing user should be redirected"""
        user = self.build_authorized_user(is_superuser=True)
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.view_url)

        self.assertRedirects(resp, "/accounts/login/?next=/user_passes_test/")

    def test_with_user_raise_exception(self):
        """PermissionDenied should be raised"""
        with self.assertRaises(PermissionDenied):
            self.dispatch_view(
                self.build_request(path=self.view_url), raise_exception=True
            )

    def test_not_implemented(self):
        """NotImplemented should be raised"""
        view = self.view_not_implemented_class()
        with self.assertRaises(NotImplementedError):
            view.dispatch(
                self.build_request(path=self.view_not_implemented_url),
                raise_exception=True,
            )


@pytest.mark.django_db
class TestSSLRequiredMixin(test.TestCase):
    """Scenarios around requiring SSL"""

    view_class = SSLRequiredView
    view_url = "/sslrequired/"

    def test_ssl_redirection(self):
        """Should redirect if not SSL"""
        self.view_url = f"https://testserver{self.view_url}"
        self.view_class.raise_exception = False
        resp = self.client.get(self.view_url)
        self.assertRedirects(resp, self.view_url, status_code=301)
        resp = self.client.get(self.view_url, follow=True)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("https", resp.request.get("wsgi.url_scheme"))

    def test_raises_exception(self):
        """Should return 404"""
        self.view_class.raise_exception = True
        resp = self.client.get(self.view_url)
        self.assertEqual(404, resp.status_code)

    @override_settings(DEBUG=True)
    def test_debug_bypasses_redirect(self):
        """Debug mode should not require SSL"""
        self.view_class.raise_exception = False
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)

    def test_https_does_not_redirect(self):
        """SSL requests should not redirect"""
        self.view_class.raise_exception = False
        resp = self.client.get(self.view_url, secure=True)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("https", resp.request.get("wsgi.url_scheme"))


@pytest.mark.django_db
class TestRecentLoginRequiredMixin(test.TestCase):
    """Scenarios requiring a recent login"""

    view_class = RecentLoginRequiredView
    recent_view_url = "/recent_login/"
    outdated_view_url = "/outdated_login/"

    def test_recent_login(self):
        """A recent login should get a 200"""
        self.view_class.max_last_login_delta = 1800
        last_login = datetime.datetime.now()
        last_login = make_aware(last_login, get_current_timezone())
        user = UserFactory(last_login=last_login)
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.recent_view_url)
        assert resp.status_code == 200
        assert force_str(resp.content) == "OK"

    def test_outdated_login(self):
        """An outdated login should get a 302"""
        self.view_class.max_last_login_delta = 0
        last_login = datetime.datetime.now() - datetime.timedelta(hours=2)
        last_login = make_aware(last_login, get_current_timezone())
        user = UserFactory(last_login=last_login)
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get(self.outdated_view_url)
        assert resp.status_code in [
            302,
            405,
        ]  # 302 is for Django < 5, while 405 is for Django >= 5

    def test_not_logged_in(self):
        """Anonymous requests should be handled appropriately"""
        last_login = datetime.datetime.now()
        last_login = make_aware(last_login, get_current_timezone())
        resp = self.client.get(self.recent_view_url)
        assert resp.status_code != 200



from django import test
from django.contrib.auth.models import AnonymousUser, User, Permission
from django.core.serializers.json import DjangoJSONEncoder

from tests.models import UserObjectPermissions


class TestViewHelper:
    """
    Helper class for unit-testing class based views.
    """

    view_class = None
    request_factory_class = test.RequestFactory

    def setUp(self):
        super().setUp()
        self.factory = self.request_factory_class()

    def build_request(self, method="GET", path="/test/", user=None, **kwargs):
        """
        Creates a request using request factory.
        """
        fn = getattr(self.factory, method.lower())
        if user is None:
            user = AnonymousUser()

        req = fn(path, **kwargs)
        req.user = user
        return req

    def build_view(
        self, request, args=None, kwargs=None, view_class=None, **viewkwargs
    ):
        """
        Creates a `view_class` view instance.
        """
        if not args:
            args = ()
        if not kwargs:
            kwargs = {}
        if view_class is None:
            view_class = self.view_class

        return view_class(
            request=request, args=args, kwargs=kwargs, **viewkwargs
        )

    def dispatch_view(
        self, request, args=None, kwargs=None, view_class=None, **viewkwargs
    ):
        """
        Creates and dispatches `view_class` view.
        """
        view = self.build_view(request, args, kwargs, view_class, **viewkwargs)
        return view.dispatch(request, *view.args, **view.kwargs)


class SetJSONEncoder(DjangoJSONEncoder):
    """
    A custom JSONEncoder extending `DjangoJSONEncoder` to handle serialization
    of `set`.
    """

    def default(self, obj):
        """Control default methods of encoding data"""
        if isinstance(obj, set):
            return list(obj)
        return super(DjangoJSONEncoder, self).default(obj)


class PermissionChecker:
    """
    Custom Permission checker for testing of Object Level Permissions
    """
    def __init__(self, user: User):
        self.user = user

    def has_perm(self, perm: str, obj=None) -> bool:
        """This function is used to check for object level permissions"""
        if self.user and not self.user.is_active:
            return False
        elif self.user and self.user.is_superuser:
            return True
        if "." in perm:
            perm = perm.split(".", maxsplit=1)[1]
        permission_obj = Permission.objects.get(codename=perm)
        if obj is None:
            return perm in self.user.get_all_permissions(perm)
        return UserObjectPermissions.objects.filter(
            permission=permission_obj,
            user=self.user,
            article_object=obj
        ).exists()

    def has_perms(self, perms: list[str], obj=None) -> bool:
        """This function is used to check for object level permissions"""
        if self.user and not self.user.is_active:
            return False
        elif self.user and self.user.is_superuser:
            return True
        if not perms:
            return False
        return all(self.has_perm(perm) for perm in perms)




from django.contrib.auth.models import Permission, User
from django.db import models


class Article(models.Model):
    """
    A small but useful model for testing most features
    """
    author = models.ForeignKey(
        "auth.User", null=True, blank=True, on_delete=models.CASCADE
    )
    title = models.CharField(max_length=30)
    body = models.TextField()
    slug = models.SlugField(blank=True)


class CanonicalArticle(models.Model):
    """
    Model specifically for testing the canonical slug mixins
    """
    author = models.ForeignKey(
        "auth.User", null=True, blank=True, on_delete=models.CASCADE
    )
    title = models.CharField(max_length=30)
    body = models.TextField()
    slug = models.SlugField(blank=True)

    def get_canonical_slug(self):
        """Required by mixin to use the model as the source of truth"""
        if self.author:
            return f"{self.author.username}-{self.slug}"
        return f"unauthored-{self.slug}"


class UserObjectPermissions(models.Model):
    """Django model used to test and assign object level permissions"""
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    permission = models.ForeignKey(Permission, on_delete=models.CASCADE)
    article_object = models.ForeignKey(Article, on_delete=models.CASCADE)




from django.contrib import messages
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.messages.storage.base import Message
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django import test
from django.test.utils import override_settings
from django.views.generic import View
from django.utils.encoding import force_str

from braces.views import (
    SetHeadlineMixin,
    MessageMixin,
    FormValidMessageMixin,
    FormInvalidMessageMixin,
)
from .factories import UserFactory
from .models import Article, CanonicalArticle
from .views import (
    FormMessagesView,
    ContextView,
)


class TestSuccessURLRedirectListMixin(test.TestCase):
    """Scenarios around redirecting after a successful form submission"""

    def test_redirect(self):
        """Successful POST should redirect"""
        data = {"title": "Test body", "body": "Test body"}
        resp = self.client.post("/article_list/create/", data)
        self.assertRedirects(resp, "/article_list/")

    def test_no_url_name(self):
        """Improper setup should raise ImproperlyConfigured"""
        data = {"title": "Test body", "body": "Test body"}
        with self.assertRaises(ImproperlyConfigured):
            # The view at this endpoint has no success_url defined
            self.client.post("/article_list_bad/create/", data)


class TestUserFormKwargsMixin(test.TestCase):
    """Scenarios around automatically including a user in form submissions"""

    def test_post_method(self):
        """A POST request should include the user kwarg"""
        user = UserFactory()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.post("/form_with_user_kwarg/", {"field1": "foo"})
        assert force_str(resp.content) == f"username: {user.username}"

    def test_get_method(self):
        """A GET request should include the user kwarg"""
        user = UserFactory()
        self.client.login(username=user.username, password="asdf1234")
        resp = self.client.get("/form_with_user_kwarg/")
        assert resp.context["form"].user == user


class TestSetHeadlineMixin(test.TestCase):
    """Scenarios around setting a headline"""

    def test_dynamic_headline(self):
        """A method-provided headline should be included in context"""
        resp = self.client.get("/headline/test-headline/")
        self.assertEqual("test-headline", resp.context["headline"])

    def test_context_data(self):
        """An attribute-provided headline should be in the context"""
        resp = self.client.get("/headline/foo-bar/")
        self.assertEqual("foo-bar", resp.context["headline"])

    def test_improper_configuration(self):
        """Not providing a headline should raise an exception"""
        mixin = SetHeadlineMixin()
        with self.assertRaises(ImproperlyConfigured):
            mixin.get_headline()

    def test_get_headline_lazy(self):
        """Lazy evaluation of text should still provide the headline"""
        resp = self.client.get("/headline/lazy/")
        self.assertEqual("Test Headline", resp.context["headline"])


class TestStaticContextMixin(test.TestCase):
    """Scenarios around including static content in the context"""

    view_class = ContextView
    view_url = "/context/"

    def test_dictionary(self):
        """Static content can be included as a dictionary"""
        self.view_class.static_context = {"test": True}
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(True, resp.context["test"])

    def test_two_tuple(self):
        """Static content can be included as a two-tuple pair"""
        self.view_class.static_context = [("a", 1), ("b", 2)]
        resp = self.client.get(self.view_url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(1, resp.context["a"])
        self.assertEqual(2, resp.context["b"])

    def test_not_set(self):
        """ImproperlyConfigured should be raised if no static content is set"""
        self.view_class.static_context = None
        with self.assertRaises(ImproperlyConfigured):
            self.client.get(self.view_url)

    def test_string_value_error(self):
        """A string should raise ImproperlyConfigured"""
        self.view_class.static_context = "Fail"
        with self.assertRaises(ImproperlyConfigured):
            self.client.get(self.view_url)

    def test_list_error(self):
        """A list should raise ImproperlyConfigured"""
        self.view_class.static_context = ["fail", "fail"]
        with self.assertRaises(ImproperlyConfigured):
            self.client.get(self.view_url)


class TestCsrfExemptMixin(test.TestCase):
    """Scenarios around views which are CSRF exempt"""

    def setUp(self):
        """Ensure the client enforces CSRF checks"""
        super(TestCsrfExemptMixin, self).setUp()
        self.client = self.client_class(enforce_csrf_checks=True)

    def test_csrf_token_is_not_required(self):
        """CSRF tokens should not be required"""
        resp = self.client.post("/csrf_exempt/", {"field1": "test"})
        self.assertEqual(200, resp.status_code)
        self.assertEqual("OK", force_str(resp.content))


class TestCanonicalSlugDetailView(test.TestCase):
    """Scenarios involving canonical slugs"""
    def setUp(self):
        """Create the two articles"""
        Article.objects.create(title="Alpha", body="Zet", slug="alpha")
        Article.objects.create(title="Zet", body="Alpha", slug="zet")

    def test_canonical_slug(self):
        """
        Test that no redirect occurs when slug is canonical.
        """
        resp = self.client.get("/article-canonical/1-alpha/")
        self.assertEqual(resp.status_code, 200)
        resp = self.client.get("/article-canonical/2-zet/")
        self.assertEqual(resp.status_code, 200)

    def test_non_canonical_slug(self):
        """
        Test that a redirect occurs when the slug is non-canonical.
        """
        resp = self.client.get("/article-canonical/1-bad-slug/")
        self.assertEqual(resp.status_code, 301)
        resp = self.client.get("/article-canonical/2-bad-slug/")
        self.assertEqual(resp.status_code, 301)


class TestNamespaceAwareCanonicalSlugDetailView(test.TestCase):
    """Scenarios around canonical slugs and namespaces"""
    def setUp(self):
        """Create the necessary articles"""
        Article.objects.create(title="Alpha", body="Zet", slug="alpha")
        Article.objects.create(title="Zet", body="Alpha", slug="zet")

    def test_canonical_slug(self):
        """
        Test that no redirect occurs when slug is canonical.
        """
        resp = self.client.get(
            "/article-canonical-namespaced/article/1-alpha/"
        )
        self.assertEqual(resp.status_code, 200)
        resp = self.client.get("/article-canonical-namespaced/article/2-zet/")
        self.assertEqual(resp.status_code, 200)

    def test_non_canonical_slug(self):
        """
        Test that a redirect occurs when the slug is non-canonical and that the
        redirect is namespace aware.
        """
        resp = self.client.get(
            "/article-canonical-namespaced/article/1-bad-slug/"
        )
        self.assertEqual(resp.status_code, 301)
        resp = self.client.get(
            "/article-canonical-namespaced/article/2-bad-slug/"
        )
        self.assertEqual(resp.status_code, 301)


class TestOverriddenCanonicalSlugDetailView(test.TestCase):
    """Scenarios involving overridden canonical slugs"""
    def setUp(self):
        """Create the necessary articles"""
        Article.objects.create(title="Alpha", body="Zet", slug="alpha")
        Article.objects.create(title="Zet", body="Alpha", slug="zet")

    def test_canonical_slug(self):
        """
        Test that no redirect occurs when slug is canonical according to the
        overridden canonical slug.
        """
        resp = self.client.get("/article-canonical-override/1-nycun/")
        self.assertEqual(resp.status_code, 200)
        resp = self.client.get("/article-canonical-override/2-mrg/")
        self.assertEqual(resp.status_code, 200)

    def test_non_canonical_slug(self):
        """
        Test that a redirect occurs when the slug is non-canonical.
        """
        resp = self.client.get("/article-canonical-override/1-bad-slug/")
        self.assertEqual(resp.status_code, 301)
        resp = self.client.get("/article-canonical-override/2-bad-slug/")
        self.assertEqual(resp.status_code, 301)


class TestCustomUrlKwargsCanonicalSlugDetailView(test.TestCase):
    """Scenarios around canonical slugs and custom URL kwargs"""
    def setUp(self):
        """Create the articles"""
        Article.objects.create(title="Alpha", body="Zet", slug="alpha")
        Article.objects.create(title="Zet", body="Alpha", slug="zet")

    def test_canonical_slug(self):
        """
        Test that no redirect occurs when slug is canonical
        """
        resp = self.client.get("/article-canonical-custom-kwargs/1-alpha/")
        self.assertEqual(resp.status_code, 200)
        resp = self.client.get("/article-canonical-custom-kwargs/2-zet/")
        self.assertEqual(resp.status_code, 200)

    def test_non_canonical_slug(self):
        """
        Test that a redirect occurs when the slug is non-canonical.
        """
        resp = self.client.get("/article-canonical-custom-kwargs/1-bad-slug/")
        self.assertEqual(resp.status_code, 301)
        resp = self.client.get("/article-canonical-custom-kwargs/2-bad-slug/")
        self.assertEqual(resp.status_code, 301)


class TestModelCanonicalSlugDetailView(test.TestCase):
    """Scenarios around canonical slugs and model fields"""
    def setUp(self):
        """Generate the necessary articles"""
        CanonicalArticle.objects.create(
            title="Alpha", body="Zet", slug="alpha"
        )
        CanonicalArticle.objects.create(title="Zet", body="Alpha", slug="zet")

    def test_canonical_slug(self):
        """
        Test that no redirect occurs when slug is canonical according to the
        model's canonical slug.
        """
        resp = self.client.get("/article-canonical-model/1-unauthored-alpha/")
        self.assertEqual(resp.status_code, 200)
        resp = self.client.get("/article-canonical-model/2-unauthored-zet/")
        self.assertEqual(resp.status_code, 200)

    def test_non_canonical_slug(self):
        """
        Test that a redirect occurs when the slug is non-canonical.
        """
        resp = self.client.get("/article-canonical-model/1-bad-slug/")
        self.assertEqual(resp.status_code, 301)
        resp = self.client.get("/article-canonical-model/2-bad-slug/")
        self.assertEqual(resp.status_code, 301)


# CookieStorage is used because it doesn't require middleware to be installed
@override_settings(
    MESSAGE_STORAGE="django.contrib.messages.storage.cookie.CookieStorage"
)
class MessageMixinTests(test.TestCase):
    """Scenarios around the messaging framework"""
    def setUp(self):
        """Create necessary objects"""
        self.rf = test.RequestFactory()
        self.middleware = MessageMiddleware("")

    def get_request(self, *args, **kwargs):
        """Generate a request that has passed through the middleware"""
        request = self.rf.get("/")
        self.middleware.process_request(request)
        return request

    def get_response(self, request, view):
        """Generate a response that has been passed through the middleware"""
        response = view(request)
        self.middleware.process_response(request, response)
        return response

    def get_request_response(self, view, *args, **kwargs):
        """Get both a request and a response, middleware-processed"""
        request = self.get_request(*args, **kwargs)
        response = self.get_response(request, view)
        return request, response

    def test_add_messages(self):
        """Message should be added through the class attribute"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.add_message(messages.SUCCESS, "test")
                return HttpResponse("OK")

        request, response = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(len(msg), 1)
        self.assertEqual(msg[0].message, "test")
        self.assertEqual(msg[0].level, messages.SUCCESS)

    def test_get_messages(self):
        """get_messages should get the stored messages"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.add_message(messages.SUCCESS, "success")
                self.messages.add_message(messages.WARNING, "warning")
                content = ",".join(
                    m.message for m in self.messages.get_messages()
                )
                return HttpResponse(content)

        _, response = self.get_request_response(TestView.as_view())
        self.assertEqual(response.content, b"success,warning")

    def test_get_level(self):
        """Should be able to get message levels"""
        class TestView(MessageMixin, View):
            def get(self, request):
                return HttpResponse(self.messages.get_level())

        _, response = self.get_request_response(TestView.as_view())
        self.assertEqual(int(response.content), messages.INFO)  # default

    def test_set_level(self):
        """Should be able to set message levels"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.set_level(messages.WARNING)
                self.messages.add_message(messages.SUCCESS, "success")
                self.messages.add_message(messages.WARNING, "warning")
                return HttpResponse("OK")

        request, _ = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(msg, [Message(messages.WARNING, "warning")])

    @override_settings(MESSAGE_LEVEL=messages.DEBUG)
    def test_debug(self):
        """Messages should able to be set as DEBUG"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.debug("test")
                return HttpResponse("OK")

        request, _ = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(len(msg), 1)
        self.assertEqual(msg[0], Message(messages.DEBUG, "test"))

    def test_info(self):
        """Messages should able to be set as INFO"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.info("test")
                return HttpResponse("OK")

        request, _ = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(len(msg), 1)
        self.assertEqual(msg[0], Message(messages.INFO, "test"))

    def test_success(self):
        """Messages should able to be set as SUCCESS"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.success("test")
                return HttpResponse("OK")

        request, _ = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(len(msg), 1)
        self.assertEqual(msg[0], Message(messages.SUCCESS, "test"))

    def test_warning(self):
        """Messages should able to be set as WARNING"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.warning("test")
                return HttpResponse("OK")

        request, _ = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(len(msg), 1)
        self.assertEqual(msg[0], Message(messages.WARNING, "test"))

    def test_error(self):
        """Messages should able to be set as ERROR"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.error("test")
                return HttpResponse("OK")

        request, _ = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(len(msg), 1)
        self.assertEqual(msg[0], Message(messages.ERROR, "test"))

    def test_invalid_attribute(self):
        """Raise an AttributeError if setting an invalid level"""
        class TestView(MessageMixin, View):
            def get(self, request):
                self.messages.invalid()
                return HttpResponse("OK")

        with self.assertRaises(AttributeError):
            self.get_request_response(TestView.as_view())

    def test_wrapper_available_in_dispatch(self):
        """
        Make sure that self.messages is available in dispatch() even before
        calling the parent's implementation.
        """

        class TestView(MessageMixin, View):
            def dispatch(self, request):
                self.messages.add_message(messages.SUCCESS, "test")
                return super(TestView, self).dispatch(request)

            def get(self, request):
                return HttpResponse("OK")

        request, response = self.get_request_response(TestView.as_view())
        msg = list(request._messages)
        self.assertEqual(len(msg), 1)
        self.assertEqual(msg[0].message, "test")
        self.assertEqual(msg[0].level, messages.SUCCESS)

    def test_API(self):
        """
        Make sure that our assumptions about messages.api are still valid.
        """
        # This test is designed to break when django.contrib.messages.api
        # changes (items being added or removed).
        excluded_API = set()
        excluded_API.add("MessageFailure")


class TestFormMessageMixins(test.TestCase):
    """Scenarios around form valid/invalid messages"""
    def setUp(self):
        self.good_data = {"title": "Good", "body": "Body"}
        self.bad_data = {"body": "Missing title"}

    def test_valid_message(self):
        """If the form is valid, the valid message should be available"""
        url = "/form_messages/"
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        response = self.client.post(url, self.good_data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, FormMessagesView().form_valid_message)

    def test_invalid_message(self):
        """If the form is invalid, the invalid message should be available"""
        url = "/form_messages/"
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        response = self.client.post(url, self.bad_data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, FormMessagesView().form_invalid_message)

    def test_form_valid_message_not_set(self):
        """Not setting a form_valid message should raise ImproperlyConfigured"""
        mixin = FormValidMessageMixin()
        with self.assertRaises(ImproperlyConfigured):
            mixin.get_form_valid_message()

    def test_form_valid_message_not_str(self):
        """Non-strings for the form_valid message should raise ImproperlyConfigured"""
        mixin = FormValidMessageMixin()
        mixin.form_valid_message = ["bad"]
        with self.assertRaises(ImproperlyConfigured):
            mixin.get_form_valid_message()

    def test_form_valid_returns_message(self):
        """get_form_valid_message should return the form_valid message"""
        mixin = FormValidMessageMixin()
        mixin.form_valid_message = "Good øø"
        self.assertEqual(force_str("Good øø"), mixin.get_form_valid_message())

    def test_form_invalid_message_not_set(self):
        """Not setting a form_invalid message should raise ImproperlyConfigured"""
        mixin = FormInvalidMessageMixin()
        with self.assertRaises(ImproperlyConfigured):
            mixin.get_form_invalid_message()

    def test_form_invalid_message_not_str(self):
        """Non-strings for the form_invalid message should raise ImproperlyConfigured"""
        mixin = FormInvalidMessageMixin()
        mixin.form_invalid_message = ["bad"]
        with self.assertRaises(ImproperlyConfigured):
            mixin.get_form_invalid_message()

    def test_form_invalid_returns_message(self):
        """get_form_invalid_message should return the form_invalid message"""
        mixin = FormInvalidMessageMixin()
        mixin.form_invalid_message = "Bad øø"
        self.assertEqual(force_str("Bad øø"), mixin.get_form_invalid_message())


class TestAllVerbsMixin(test.TestCase):
    """Scenarios around the AllVerbsMixin"""
    def setUp(self):
        self.url = "/all_verbs/"
        self.no_handler_url = "/all_verbs_no_handler/"

    def test_options(self):
        """AllVerbs should respond to OPTION"""
        response = self.client.options(self.url)
        self.assertEqual(response.status_code, 200)

    def test_get(self):
        """AllVerbs should respond to GET"""
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

    def test_head(self):
        """AllVerbs should respond to HEAD"""
        response = self.client.head(self.url)
        self.assertEqual(response.status_code, 200)

    def test_post(self):
        """AllVerbs should respond to POST"""
        response = self.client.post(self.url)
        self.assertEqual(response.status_code, 200)

    def test_put(self):
        """AllVerbs should respond to PUT"""
        response = self.client.put(self.url)
        self.assertEqual(response.status_code, 200)

    def test_delete(self):
        """AllVerbs should respond to DELETE"""
        response = self.client.delete(self.url)
        self.assertEqual(response.status_code, 200)

    def test_patch(self):
        """AllVerbs should respond to PATCH"""
        response = self.client.patch(self.url)
        self.assertEqual(response.status_code, 200)

    def test_no_all_handler(self):
        """A missing handler should raise ImproperlyConfigured"""
        with self.assertRaises(ImproperlyConfigured):
            self.client.get("/all_verbs_no_handler/")


class TestHeaderMixin(test.TestCase):
    """Scenarios around the extra headers mixin"""
    def test_attribute(self):
        """Headers can be set via an attribute"""
        response = self.client.get("/headers/attribute/")
        self.assertEqual(response["X-DJANGO-BRACES-1"], "1")
        self.assertEqual(response["X-DJANGO-BRACES-2"], "2")

    def test_method(self):
        """Headers can be set via a method"""
        response = self.client.get("/headers/method/")
        self.assertEqual(response["X-DJANGO-BRACES-1"], "1")
        self.assertEqual(response["X-DJANGO-BRACES-2"], "2")

    def test_existing(self):
        """Existing headers should still come through"""
        response = self.client.get('/headers/existing/')
        self.assertEqual(response['X-DJANGO-BRACES-EXISTING'], 'value')

class TestCacheControlMixin(test.TestCase):
    """Scenarios around controlling cache"""
    def test_cachecontrol_public(self):
        """Cache settings should be respected and included"""
        response = self.client.get('/cachecontrol/public/')
        options = [i.strip() for i in response['Cache-Control'].split(',')]
        self.assertEqual(sorted(options), ['max-age=60', 'public'])


class TestNeverCacheMixin(test.TestCase):
    """Scenarios around marking a view as never-cached"""
    def test_nevercache(self):
        """Views marked as no-cache should not be cached"""
        response = self.client.get('/nevercache/')
        options = [i.strip() for i in response['Cache-Control'].split(',')]
        expected_cache_control_options = {"max-age=0", "must-revalidate", "no-cache", "no-store", "private"}
        self.assertTrue(set(options).intersection(expected_cache_control_options))
